import base64
import gzip
from datetime import datetime, date, time  # noqa: I251
from pendulum.tz import UTC
from typing import (
    Any,
    Dict,
    Mapping,
    NamedTuple,
    Tuple,
    Optional,
    Union,
    Callable,
    Iterable,
    Iterator,
    Sequence,
    List,
)

from dlt import version
from dlt.common.exceptions import MissingDependencyException, DltException
from dlt.common.schema.typing import C_DLT_ID, C_DLT_LOAD_ID, TColumnSchema, TTableSchemaColumns
from dlt.common import logger
from dlt.common.json import json, custom_encode, map_nested_values_in_place
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.schema.typing import TColumnType
from dlt.common.schema.utils import is_nullable_column, dlt_load_id_column
from dlt.common.time import get_precision_from_datetime_unit
from dlt.common.typing import AnyType, StrStr, TFileOrPath, TDataItems
from dlt.common.normalizers.naming import NamingConvention

try:
    import pyarrow
    import pyarrow.parquet
    import pyarrow.compute
    import pyarrow.dataset
    from pyarrow.parquet import ParquetFile
    from pyarrow import Table
except ModuleNotFoundError:
    raise MissingDependencyException(
        "dlt pyarrow helpers",
        [f"{version.DLT_PKG_NAME}[parquet]"],
        "Install pyarrow to be allowed to load arrow tables, pandas DataFrames and to use parquet"
        " files.",
    )

import ctypes

TAnyArrowItem = Union[pyarrow.Table, pyarrow.RecordBatch]

ARROW_DECIMAL_MAX_PRECISION = 76


class UnsupportedArrowTypeException(DltException):
    """Exception raised when Arrow type conversion failed.

    The setters are used to update the exception with more context
    such as the relevant field and tablea it is caught downstream.
    """

    def __init__(
        self,
        arrow_type: pyarrow.DataType,
        field_name: Optional[str] = None,
        table_name: Optional[str] = None,
    ) -> None:
        self.arrow_type = arrow_type
        self._field_name = field_name if field_name else ""
        self._table_name = table_name if table_name else ""

        msg = self.generate_message(self.arrow_type, self._field_name, self._table_name)
        super().__init__(msg)

    @staticmethod
    def generate_message(arrow_type: pyarrow.DataType, field_name: str, table_name: str) -> str:
        msg = f"Arrow type `{arrow_type}`"
        if field_name:
            msg += f" for field `{field_name}`"
        if table_name:
            msg += f" in table `{table_name}`"

        msg += (
            " is unsupported by dlt. See documentation:"
            " https://dlthub.com/docs/dlt-ecosystem/verified-sources/arrow-pandas#supported-arrow-data-types"
        )
        return msg

    def _update_message(self) -> None:
        """Modify the `Exception.args` tuple to update message."""
        msg = self.generate_message(self.arrow_type, self.field_name, self.table_name)
        self.args = (msg,)  # must be a tuple

    @property
    def field_name(self) -> str:
        return self._field_name

    @field_name.setter
    def field_name(self, value: str) -> None:
        self._field_name = value
        self._update_message()

    @property
    def table_name(self) -> str:
        return self._table_name

    @table_name.setter
    def table_name(self, value: str) -> None:
        self._table_name = value
        self._update_message()


class PyToArrowConversionException(DltException):
    """Exception raised when converting data to Arrow based on a TableSchema"""

    def __init__(
        self,
        data_type: Optional[str],
        inferred_arrow_type: Optional[pyarrow.DataType] = None,
        field_name: Optional[str] = None,
        table_name: Optional[str] = None,
        details: Optional[str] = None,
    ) -> None:
        self.data_type = data_type
        self.inferred_arrow_type = inferred_arrow_type
        self._field_name = field_name if field_name else ""
        self._table_name = table_name if table_name else ""
        self._details = details if details else ""

        super().__init__()
        self._update_message()

    @staticmethod
    def generate_message(
        data_type: Optional[str],
        inferred_arrow_type: Optional[pyarrow.DataType],
        field_name: str,
        table_name: str,
        details: str,
    ) -> str:
        msg = "Conversion to arrow failed"
        if field_name:
            msg += f" for field `{field_name}`"
        if table_name:
            msg += f" in table `{table_name}`"

        msg += f" with dlt hint `{data_type=:}` and `{inferred_arrow_type=:}`"
        msg += " " + details
        return msg

    def _update_message(self) -> None:
        """Modify the `Exception.args` tuple to update message."""
        msg = self.generate_message(
            self.data_type,
            self.inferred_arrow_type,
            self.field_name,
            self.table_name,
            self._details,
        )
        self.args = (msg,)  # must be a tuple

    @property
    def field_name(self) -> str:
        return self._field_name

    @field_name.setter
    def field_name(self, value: str) -> None:
        self._field_name = value
        self._update_message()

    @property
    def table_name(self) -> str:
        return self._table_name

    @table_name.setter
    def table_name(self, value: str) -> None:
        self._table_name = value
        self._update_message()

    @property
    def details(self) -> str:
        return self._details

    @details.setter
    def details(self, value: str) -> None:
        self._details = value
        self._update_message()


class ArrowSchemaNormalizationResult(NamedTuple):
    """Named result of should_normalize_arrow_schema

    Fields:
    - should_normalize: whether any normalization is required
    - rename_mapping: mapping from original field names to normalized names
    - rev_mapping: reverse mapping from normalized names to original
    - nullable_updates: fields that require nullable flag updates (by normalized name)
    - columns: potentially filtered/adjusted TTableSchemaColumns
    """

    should_normalize: bool
    rename_mapping: StrStr
    rev_mapping: Dict[str, str]
    nullable_updates: Dict[str, bool]
    columns: TTableSchemaColumns


def get_py_arrow_datatype(
    column: TColumnType,
    caps: DestinationCapabilitiesContext,
    tz: str,
) -> Any:
    column_type = column["data_type"]
    if column_type == "text":
        return pyarrow.string()
    elif column_type == "double":
        return pyarrow.float64()
    elif column_type == "bool":
        return pyarrow.bool_()
    elif column_type == "timestamp":
        # sets timezone to None when timezone hint is false
        timezone = tz if column.get("timezone", True) else None
        precision = column.get("precision")
        if precision is None:
            precision = caps.timestamp_precision
        return get_py_arrow_timestamp(precision, timezone)
    elif column_type == "bigint":
        return get_pyarrow_int(column.get("precision"))
    elif column_type == "binary":
        return pyarrow.binary(column.get("precision") or -1)
    elif column_type == "json":
        if (nested_type := column.get("x-nested-type")) and caps.supports_nested_types:
            return deserialize_type(nested_type)  # type: ignore[arg-type]
        else:
            return pyarrow.string()
    elif column_type == "decimal":
        precision, scale = column.get("precision"), column.get("scale")
        if (precision is None) and (scale is None):
            precision_tuple = caps.decimal_precision
        elif precision is None:
            precision_tuple = caps.decimal_precision
            logger.warning(
                f"Received decimal column hint `scale={scale}`, but `precision` not set. Will"
                " assume default destination capability `(precision, scale) ="
                f" {caps.decimal_precision}`"
            )
        elif scale is None:
            # setting scale to 0 when unspecified is a common practice across databases
            precision_tuple = (precision, 0)
            logger.warning(
                f"Received decimal column hint `precision={precision}`, but `scale` not set. "
                "Will assume default destination capability `scale=0`"
            )
        else:
            precision_tuple = (precision, scale)

        return get_py_arrow_numeric(precision_tuple)
    elif column_type == "wei":
        return get_py_arrow_numeric(caps.wei_precision)
    elif column_type == "date":
        return pyarrow.date32()
    elif column_type == "time":
        precision = column.get("precision")
        if precision is None:
            precision = caps.timestamp_precision
        return get_py_arrow_time(precision)
    else:
        raise ValueError(column_type)


def get_py_arrow_timestamp(precision: int, tz: str) -> Any:
    tz = tz if tz else None
    if precision == 0:
        return pyarrow.timestamp("s", tz=tz)
    if precision <= 3:
        return pyarrow.timestamp("ms", tz=tz)
    if precision <= 6:
        return pyarrow.timestamp("us", tz=tz)
    return pyarrow.timestamp("ns", tz=tz)


def get_py_arrow_time(precision: int) -> Any:
    if precision == 0:
        return pyarrow.time32("s")
    elif precision <= 3:
        return pyarrow.time32("ms")
    elif precision <= 6:
        return pyarrow.time64("us")
    return pyarrow.time64("ns")


def get_py_arrow_numeric(precision: Tuple[int, int]) -> Any:
    if precision[0] <= 38:
        return pyarrow.decimal128(*precision)
    if precision[0] <= 76:
        return pyarrow.decimal256(*precision)
    # for higher precision use max precision and trim scale to leave the most significant part
    return pyarrow.decimal256(76, max(0, 76 - (precision[0] - precision[1])))


def get_pyarrow_int(precision: Optional[int]) -> Any:
    if precision is None:
        return pyarrow.int64()
    if precision <= 8:
        return pyarrow.int8()
    elif precision <= 16:
        return pyarrow.int16()
    elif precision <= 32:
        return pyarrow.int32()
    return pyarrow.int64()


def get_column_type_from_py_arrow(dtype: pyarrow.DataType) -> TColumnType:
    """Returns (data_type, precision, scale) tuple from pyarrow.DataType"""
    if pyarrow.types.is_string(dtype) or pyarrow.types.is_large_string(dtype):
        return dict(data_type="text")
    elif pyarrow.types.is_floating(dtype):
        return dict(data_type="double")
    elif pyarrow.types.is_boolean(dtype):
        return dict(data_type="bool")
    elif pyarrow.types.is_timestamp(dtype):
        precision = get_precision_from_datetime_unit(dtype.unit)
        timestamp_d: TColumnType = dict(data_type="timestamp", precision=precision)
        if dtype.tz is None:
            timestamp_d["timezone"] = False
        return timestamp_d
    elif pyarrow.types.is_date(dtype):
        return dict(data_type="date")
    elif pyarrow.types.is_time(dtype):
        # Time fields in schema are `DataType` instead of `Time64Type` or `Time32Type`
        precision = get_precision_from_datetime_unit(dtype.unit)
        return dict(data_type="time", precision=precision)
    elif pyarrow.types.is_integer(dtype):
        result: TColumnType = dict(data_type="bigint")
        if dtype.bit_width != 64:  # 64bit is a default bigint
            result["precision"] = dtype.bit_width
        return result
    elif pyarrow.types.is_fixed_size_binary(dtype):
        return dict(data_type="binary", precision=dtype.byte_width)
    elif pyarrow.types.is_binary(dtype) or pyarrow.types.is_large_binary(dtype):
        return dict(data_type="binary")
    elif pyarrow.types.is_decimal(dtype):
        return dict(data_type="decimal", precision=dtype.precision, scale=dtype.scale)
    elif pyarrow.types.is_nested(dtype):
        return get_nested_column_type_from_py_arrow(dtype)
    elif pyarrow.types.is_dictionary(dtype):
        # Dictionary types are essentially categorical encodings. The underlying value_type
        # dictates the "logical" type. We simply delegate to the underlying value_type.
        return get_column_type_from_py_arrow(dtype.value_type)
    elif pyarrow.types.is_null(dtype):
        return {}  # incomplete column, no data_type
    else:
        raise UnsupportedArrowTypeException(arrow_type=dtype)


def py_arrow_to_table_schema_columns(schema: pyarrow.Schema) -> TTableSchemaColumns:
    """Convert a PyArrow schema to a table schema columns dict.

    Args:
        schema (pyarrow.Schema): pyarrow schema

    Returns:
        TTableSchemaColumns: table schema columns
    """
    result: TTableSchemaColumns = {}
    for field in schema:
        try:
            converted_type = get_column_type_from_py_arrow(field.type)
        except UnsupportedArrowTypeException as e:
            # modify attributes inplace to add context instead of re-raising with `raise e`
            e.field_name = field.name
            raise

        result[field.name] = {
            "name": field.name,
            "nullable": field.nullable,
            **converted_type,
        }
    return result


def get_nested_column_type_from_py_arrow(dtype: pyarrow.DataType) -> TColumnType:
    """Creates `json` dlt data type with nested type structure in `x-nested-type` hint.
    Currently the only recognized nested type format is arrow-ipc
    """
    return {"data_type": "json", "x-nested-type": serialize_type(dtype)}  # type: ignore[typeddict-unknown-key]


def serialize_type(dtype: pyarrow.DataType) -> str:
    """Serializes arrow type via arrow ipc as base64 str"""
    schema = pyarrow.schema([pyarrow.field("c", dtype)])
    return "arrow-ipc:" + base64.b64encode(gzip.compress(schema.serialize().to_pybytes())).decode(
        "ascii"
    )


def deserialize_type(type_str: str) -> pyarrow.DataType:
    if type_str.startswith("arrow-ipc:"):
        decompressed = gzip.decompress(base64.b64decode(type_str[10:]))
        schema = pyarrow.ipc.read_schema(pyarrow.BufferReader(decompressed))
        return schema.field(0).type
    else:
        raise TypeError("Cannot deserialize pyarrow type, only arrow-ipc is supported")


def remove_null_columns(item: TAnyArrowItem) -> TAnyArrowItem:
    """Remove all columns of datatype pyarrow.null() from the table or record batch.
    Stores removed column names in arrow schema metadata under 'dlt.null_columns' key.
    """
    null_col_names = [field.name for field in item.schema if pyarrow.types.is_null(field.type)]
    if not null_col_names:
        return item
    item = remove_columns(item, null_col_names)
    return add_arrow_metadata(item, {"dlt.null_columns": json.dumps(null_col_names)})


def remove_null_columns_from_schema(schema: pyarrow.Schema) -> Tuple[pyarrow.Schema, bool]:
    """Remove all columns of datatype pyarrow.null() from the schema"""
    fields: List[pyarrow.field] = []
    contains_null: bool = False
    for field in schema:
        if pyarrow.types.is_null(field.type):
            contains_null = True
        else:
            fields.append(field)
    return pyarrow.schema(fields), contains_null


def remove_columns(item: TAnyArrowItem, columns: Sequence[str]) -> TAnyArrowItem:
    """Remove `columns` from Arrow `item`"""
    if not columns:
        return item

    if isinstance(item, pyarrow.Table):
        return item.drop(columns)
    elif isinstance(item, pyarrow.RecordBatch):
        # NOTE: select is available in pyarrow 12 an up
        return item.select([n for n in item.schema.names if n not in columns])  # reverse selection
    else:
        raise ValueError(item)


def append_column(item: TAnyArrowItem, name: str, data: Any) -> TAnyArrowItem:
    """Appends new column to Table or RecordBatch"""
    if isinstance(item, pyarrow.Table):
        return item.append_column(name, data)
    elif isinstance(item, pyarrow.RecordBatch):
        new_field = pyarrow.field(name, data.type)
        return pyarrow.RecordBatch.from_arrays(
            item.columns + [data], schema=item.schema.append(new_field)
        )
    else:
        raise ValueError(item)


def rename_columns(item: TAnyArrowItem, new_column_names: Sequence[str]) -> TAnyArrowItem:
    """Rename arrow columns on Table or RecordBatch, returns same data but with renamed schema"""

    if list(item.schema.names) == list(new_column_names):
        # No need to rename
        return item

    if isinstance(item, pyarrow.Table):
        return item.rename_columns(new_column_names)
    elif isinstance(item, pyarrow.RecordBatch):
        new_fields = [
            field.with_name(new_name) for new_name, field in zip(new_column_names, item.schema)
        ]
        return pyarrow.RecordBatch.from_arrays(item.columns, schema=pyarrow.schema(new_fields))
    else:
        raise TypeError(f"Unsupported data item type: `{type(item)}`")


def fill_empty_source_column_values_with_placeholder(
    table: pyarrow.Table, source_columns: List[str], placeholder: str
) -> pyarrow.Table:
    """
    Replaces empty strings and null values in the specified source columns of an Arrow table with a placeholder string.

    Args:
        table (pa.Table): The input Arrow table.
        source_columns (List[str]): A list of column names to replace empty strings and null values in.
        placeholder (str): The placeholder string to use for replacement.

    Returns:
        pyarrow.Table: The modified Arrow table with empty strings and null values replaced in the specified columns.
    """
    for col_name in source_columns:
        column = table[col_name]
        filled_column = pyarrow.compute.fill_null(column, fill_value=placeholder)
        new_column = pyarrow.compute.replace_substring_regex(
            filled_column, pattern=r"^$", replacement=placeholder
        )
        table = table.set_column(table.column_names.index(col_name), col_name, new_column)
    return table


def should_normalize_arrow_schema(
    schema: pyarrow.Schema,
    columns: TTableSchemaColumns,
    naming: NamingConvention,
) -> ArrowSchemaNormalizationResult:
    """Figure out if any of the normalization steps must be executed. This prevents
    from rewriting arrow tables when no changes are needed. Refer to `normalize_py_arrow_item`
    for a list of normalizations. Note that `column` must be already normalized.
    """
    schema, contains_null_cols = remove_null_columns_from_schema(schema)

    rename_mapping = get_normalized_arrow_fields_mapping(schema, naming)
    # no clashes in rename ensured above
    rev_mapping = {v: k for k, v in rename_mapping.items()}
    nullable_mapping = {k: is_nullable_column(v) for k, v in columns.items()}

    # Key is the renamed column name
    nullable_updates: Dict[str, bool] = {}
    column_cast: Dict[str, bool] = {}
    for field in schema:
        norm_name = rename_mapping[field.name]
        # All fields from arrow schema that have nullable set to different value than in columns
        if norm_name in nullable_mapping and field.nullable != nullable_mapping[norm_name]:
            nullable_updates[norm_name] = nullable_mapping[norm_name]
        # Detect arrow columns that require to be normalized
        if norm_name in columns and should_normalize_py_arrow_item_column(
            columns[norm_name], field.type
        ):
            column_cast[norm_name] = True

    dlt_load_id_col = naming.normalize_identifier(C_DLT_LOAD_ID)
    dlt_id_col = naming.normalize_identifier(C_DLT_ID)
    dlt_columns = {dlt_load_id_col, dlt_id_col}

    # remove all columns that are dlt columns but are not present in arrow schema. we do not want to add such columns
    # that should happen in the normalizer
    columns = {
        name: column
        for name, column in columns.items()
        if name not in dlt_columns or name in rev_mapping
    }

    # check if nothing to rename
    skip_normalize = (
        (list(rename_mapping.keys()) == list(rename_mapping.values()) == list(columns.keys()))
        and not nullable_updates
        and not contains_null_cols
        and not column_cast
    )
    return ArrowSchemaNormalizationResult(
        not skip_normalize,
        rename_mapping,
        rev_mapping,
        nullable_updates,
        columns,
    )


def normalize_py_arrow_item(
    item: TAnyArrowItem,
    columns: TTableSchemaColumns,
    naming: NamingConvention,
    caps: DestinationCapabilitiesContext,
) -> TAnyArrowItem:
    """Normalize arrow `item` schema according to the `columns`. Note that
    columns must be already normalized.

    0. columns with no data type will be dropped
    1. arrow schema field names will be normalized according to `naming`
    2. arrows columns will be reordered according to `columns`
    3. empty columns will be inserted if they are missing, types will be generated using `caps`
    4. arrow columns with different nullability than corresponding schema columns will be updated
    5. timestamps will be normalized according to timezone flag in dlt column schema

    NOTE: nullability is not enforced. it is up to destination to do that.
    """
    item = remove_null_columns(item)
    schema = item.schema
    should_normalize, rename_mapping, rev_mapping, nullable_updates, columns = (
        should_normalize_arrow_schema(schema, columns, naming)
    )
    if not should_normalize:
        return item

    new_fields = []
    new_columns = []

    for column_name, column in columns.items():
        # get original field name
        field_name = rev_mapping.pop(column_name, column_name)
        if field_name in rename_mapping:
            idx = schema.get_field_index(field_name)
            new_field = schema.field(idx).with_name(column_name)
            if column_name in nullable_updates:
                # Set field nullable to match column
                new_field = new_field.with_nullable(nullable_updates[column_name])

            # coerce type
            new_type, new_arrow_column = normalize_py_arrow_item_column(
                column, new_field.type, item.column(idx)
            )

            # use renamed field
            new_fields.append(new_field.with_type(new_type))
            new_columns.append(new_arrow_column)
        else:
            # column does not exist in pyarrow. create empty field and column
            new_field = pyarrow.field(
                column_name,
                get_py_arrow_datatype(column, caps, "UTC"),
                nullable=is_nullable_column(column),
            )
            new_fields.append(new_field)
            new_columns.append(pyarrow.nulls(item.num_rows, type=new_field.type))

    # add the remaining columns
    for column_name, field_name in rev_mapping.items():
        idx = schema.get_field_index(field_name)
        # use renamed field
        new_fields.append(schema.field(idx).with_name(column_name))
        new_columns.append(item.column(idx))

    # preserve schema metadata (e.g. dlt.null_columns) through normalization rebuild
    return item.__class__.from_arrays(
        new_columns, schema=pyarrow.schema(new_fields, metadata=item.schema.metadata)
    )


def should_normalize_py_arrow_item_column(
    column: TColumnSchema, arrow_type: pyarrow.DataType
) -> bool:
    # Only handle timestamp columns
    if not pyarrow.types.is_timestamp(arrow_type):
        return False

    current_tz = arrow_type.tz
    target_tz = "UTC" if column.get("timezone", True) else None

    # normalize if tz different
    return current_tz != target_tz  # type: ignore[no-any-return]


def normalize_py_arrow_item_column(
    column: TColumnSchema, arrow_type: pyarrow.Field, arrow_column: pyarrow.Array
) -> Tuple[pyarrow.DataType, pyarrow.Array]:
    """Normalize arrow timestamp column timezone according to dlt schema column convention.

    Args:
        column: dlt column schema with timezone hint
        arrow_type: actual PyArrow data type
        arrow_column: actual PyArrow column data

    Returns:
        Tuple of (modified_type, modified_column) or (arrow_field, arrow_column) if no changes needed
    """
    if not should_normalize_py_arrow_item_column(column, arrow_type):
        return arrow_type, arrow_column

    unit = arrow_type.unit
    current_tz = arrow_type.tz
    target_tz = "UTC" if column.get("timezone", True) else None

    if target_tz == "UTC":
        # Need tz-aware UTC
        if current_tz is None:
            # Attach UTC without shifting (values already represent UTC)
            col = pyarrow.compute.assume_timezone(
                arrow_column, "UTC", ambiguous="latest", nonexistent="latest"
            )
        else:
            # Metadata-only cast to UTC tz
            col = pyarrow.compute.cast(arrow_column, pyarrow.timestamp(unit, "UTC"))
    else:
        # Need naive
        if current_tz is None:
            col = arrow_column  # already naive
        else:
            # Make naive wall-clock; if you want naive-in-UTC, ensure tz metadata is UTC first
            if current_tz != "UTC":
                arrow_column = pyarrow.compute.cast(arrow_column, pyarrow.timestamp(unit, "UTC"))
            col = pyarrow.compute.local_timestamp(arrow_column)

    return pyarrow.timestamp(unit, target_tz), col


def add_dlt_load_id_column(
    item: TAnyArrowItem,
    columns: TTableSchemaColumns,
    caps: DestinationCapabilitiesContext,
    naming: NamingConvention,
    load_id: str,
) -> TAnyArrowItem:
    """
    Adds or replaces the `_dlt_load_id` column.
    """
    dlt_load_id_col_name = naming.normalize_identifier(C_DLT_LOAD_ID)

    idx = item.schema.get_field_index(dlt_load_id_col_name)
    # if the column already exists, get rid of it
    if idx != -1:
        item = remove_columns(item, dlt_load_id_col_name)

    # get pyarrow.string() type
    pyarrow_string = get_py_arrow_datatype(
        # use already existing column definition or use the default
        # NOTE: the existence of the load id column is ensured by this time
        # since it is added in _compute_tables before files are written
        (
            columns[dlt_load_id_col_name]
            if dlt_load_id_col_name in columns
            else dlt_load_id_column()
        ),
        caps,
        "UTC",  # ts is irrelevant to get pyarrow string, but it's required...
    )

    # Check if destination supports dictionary encoding (default True if not specified)
    use_dictionary = True
    if caps.parquet_format is not None:
        use_dictionary = caps.parquet_format.supports_dictionary_encoding

    # add the column with the new value at previous index or append
    item = add_constant_column(
        item=item,
        name=dlt_load_id_col_name,
        data_type=pyarrow_string,
        value=load_id,
        nullable=(
            columns[dlt_load_id_col_name]["nullable"]
            if dlt_load_id_col_name in columns
            else dlt_load_id_column()["nullable"]
        ),
        index=idx,
        use_dictionary=use_dictionary,
    )

    return item


def get_normalized_arrow_fields_mapping(schema: pyarrow.Schema, naming: NamingConvention) -> StrStr:
    """Normalizes schema field names and returns mapping from original to normalized name. Raises on name collisions"""
    # use normalize_path to be compatible with how regular columns are normalized in dlt.Schema
    norm_f = naming.normalize_path
    name_mapping = {n.name: norm_f(n.name) for n in schema}
    # verify if names uniquely normalize
    normalized_names = set(name_mapping.values())
    if len(name_mapping) != len(normalized_names):
        raise NameNormalizationCollision(
            f"Arrow schema fields normalized from:\n{list(name_mapping.keys())}:\nto:\n"
            f" {list(normalized_names)}"
        )
    return name_mapping


def columns_to_arrow(
    columns: TTableSchemaColumns,
    caps: DestinationCapabilitiesContext,
    timestamp_timezone: str = "UTC",
) -> pyarrow.Schema:
    """Convert a table schema columns dict to a pyarrow schema.

    Args:
        columns (TTableSchemaColumns): table schema columns

    Returns:
        pyarrow.Schema: pyarrow schema

    """
    caps = caps or DestinationCapabilitiesContext.generic_capabilities()
    return pyarrow.schema(
        [
            pyarrow.field(
                name,
                get_py_arrow_datatype(
                    schema_item,
                    caps,
                    timestamp_timezone,
                ),
                nullable=schema_item.get("nullable", True),
            )
            for name, schema_item in columns.items()
            if schema_item.get("data_type") is not None
        ]
    )


def get_parquet_metadata(parquet_file: TFileOrPath) -> Tuple[int, pyarrow.Schema]:
    """Gets parquet file metadata (including row count and schema)

    Args:
        parquet_file (str): path to parquet file

    Returns:
        FileMetaData: file metadata
    """
    with pyarrow.parquet.ParquetFile(parquet_file) as reader:
        return reader.metadata.num_rows, reader.schema_arrow


def is_arrow_item(item: Any) -> bool:
    return isinstance(item, (pyarrow.Table, pyarrow.RecordBatch))


def to_arrow_scalar(value: Any, arrow_type: pyarrow.DataType) -> Any:
    """Converts python value to an arrow compute friendly version"""
    return pyarrow.scalar(value, type=arrow_type)


def from_arrow_scalar(arrow_value: pyarrow.Scalar) -> Any:
    """Converts arrow scalar into Python type."""
    return arrow_value.as_py()


TNewColumns = Sequence[Tuple[int, pyarrow.Field, Callable[[pyarrow.Table], Iterable[Any]]]]
"""Sequence of tuples: (field index, field, generating function)"""


def add_constant_column(
    item: TAnyArrowItem,
    name: str,
    data_type: pyarrow.DataType,
    value: Any = None,
    nullable: bool = True,
    index: int = -1,
    use_dictionary: bool = True,
) -> TAnyArrowItem:
    """Add column with a single value to the table.

    Args:
        item: Arrow table or record batch
        name: The new column name
        data_type: The data type of the new column
        nullable: Whether the new column is nullable
        value: The value to fill the new column with
        index: The index at which to insert the new column. Defaults to -1 (append)
        use_dictionary: When True (default), creates a dictionary-encoded column which is
            memory-efficient for repeated values. Set to False for destinations that don't
            support dictionary types (e.g., ADBC drivers for MSSQL).
    Note:
        When use_dictionary=True, the column is created as a DictionaryArray with int8 indices.
        When use_dictionary=False, a regular array filled with the repeated value is created.
    """
    if use_dictionary:
        dictionary = pyarrow.array([value], type=data_type)
        zero_buffer = pyarrow.allocate_buffer(item.num_rows, resizable=False)
        ctypes.memset(zero_buffer.address, 0, item.num_rows)

        indices = pyarrow.Array.from_buffers(
            pyarrow.int8(),
            item.num_rows,
            [None, zero_buffer],  # None validity bitmap means arrow assumes all entries are valid
        )
        column_array = pyarrow.DictionaryArray.from_arrays(indices, dictionary)
    else:
        # Create a regular array filled with the repeated value
        column_array = pyarrow.repeat(pyarrow.scalar(value, type=data_type), item.num_rows)

    field = pyarrow.field(name, column_array.type, nullable=nullable)
    if index == -1:
        return item.append_column(field, column_array)
    return item.add_column(index, field, column_array)


def pq_stream_with_new_columns(
    parquet_file: TFileOrPath, columns: TNewColumns, row_groups_per_read: int = 1
) -> Iterator[pyarrow.Table]:
    """Add column(s) to the table in batches.

    The table is read from parquet `row_groups_per_read` row groups at a time

    Args:
        parquet_file: path or file object to parquet file
        columns: list of columns to add in the form of (insertion index, `pyarrow.Field`, column_value_callback)
            The callback should accept a `pyarrow.Table` and return an array of values for the column.
        row_groups_per_read: number of row groups to read at a time. Defaults to 1.

    Yields:
        `pyarrow.Table` objects with the new columns added.
    """
    with pyarrow.parquet.ParquetFile(parquet_file) as reader:
        n_groups = reader.num_row_groups
        # Iterate through n row groups at a time
        for i in range(0, n_groups, row_groups_per_read):
            tbl: pyarrow.Table = reader.read_row_groups(
                range(i, min(i + row_groups_per_read, n_groups))
            )
            for idx, field, gen_ in columns:
                if idx == -1:
                    tbl = tbl.append_column(field, gen_(tbl))
                else:
                    tbl = tbl.add_column(idx, field, gen_(tbl))
            yield tbl


def cast_arrow_schema_types(
    schema: pyarrow.Schema,
    type_map: Dict[Callable[[pyarrow.DataType], bool], Callable[..., pyarrow.DataType]],
) -> pyarrow.Schema:
    """Returns type-casted Arrow schema.

    Replaces data types for fields matching a type check in `type_map`.
    Type check functions in `type_map` are assumed to be mutually exclusive, i.e.
    a data type does not match more than one type check function.
    """
    for i, e in enumerate(schema.types):
        for type_check, cast_type in type_map.items():
            if type_check(e):
                if callable(cast_type):
                    cast_type = cast_type(e)
                adjusted_field = schema.field(i).with_type(cast_type)
                schema = schema.set(i, adjusted_field)
                break  # if type matches type check, do not do other type checks
    return schema


def concat_batches_and_tables_in_order(
    tables_or_batches: Iterable[Union[pyarrow.Table, pyarrow.RecordBatch]]
) -> pyarrow.Table:
    """Concatenate iterable of tables and batches into a single table, preserving row order. Zero copy is used during
    concatenation so schemas must be identical.
    """
    batches = []
    tables = []
    for item in tables_or_batches:
        if isinstance(item, pyarrow.RecordBatch):
            batches.append(item)
        elif isinstance(item, pyarrow.Table):
            if batches:
                tables.append(pyarrow.Table.from_batches(batches))
                batches = []
            tables.append(item)
        else:
            raise ValueError(f"Unsupported type: `{type(item)}`")
    if batches:
        tables.append(pyarrow.Table.from_batches(batches))
    # "none" option ensures 0 copy concat
    return pyarrow.concat_tables(tables, promote_options="none")


def transpose_rows_to_columns(
    rows: TDataItems, column_names: Iterable[str]
) -> dict[str, Any]:  # dict[str, np.ndarray]
    """Transpose rows (data items) into columns (numpy arrays). Returns a dictionary of {column_name: column_data}

    Uses pandas if available. Otherwise, use numpy, which is slower
    """
    try:
        from dlt.common.libs.numpy import numpy as np
    except MissingDependencyException:
        raise MissingDependencyException(
            "dlt pyarrow helpers", ["numpy"], "Numpy is required for this pyarrow operation"
        )

    try:
        from pandas._libs import lib

        # NOTE: this is part of public interface now via DataFrame.from_records()
        pivoted_rows = lib.to_object_array_tuples(rows).T
    except ImportError:
        logger.info(
            "Pandas not installed, reverting to numpy.asarray to create a table which is slower"
        )
        pivoted_rows = np.asarray(rows, dtype="object", order="K").T
    return {
        column_name: data.ravel()
        for column_name, data in zip(column_names, np.vsplit(pivoted_rows, len(pivoted_rows)))
    }


def convert_numpy_to_arrow(
    column_data: Any,  # 1-dimensional np.ndarray
    caps: DestinationCapabilitiesContext,
    column_schema: TColumnSchema,
    tz: str,
    safe_arrow_conversion: bool,
) -> Any:  # pyarrow.Array
    """Convert a numpy array to a pyarrow array.

    Args:
        rows: data items
        caps: capabilities of the storage backend
        columns: dlt hints about the table columns (e.g., data type, nullabe)
        tz: time zone identifier
        safe_arrow_conversion: if False, truncation and loss of precision is allowed
            ref: https://arrow.apache.org/docs/python/generated/pyarrow.compute.CastOptions.html#pyarrow.compute.CastOptions

    Returns:
        an arrow Array
    """
    from dlt.common.libs.pyarrow import pyarrow as pa

    dlt_data_type = column_schema.get("data_type")
    inferred_arrow_type = (
        get_py_arrow_datatype(column_schema, caps, tz) if dlt_data_type is not None else None
    )
    inferred_array = None

    # base case (0): allow pyarrow to infer type, or create array of dlt specified type
    try:
        # type=None lets pyarrow infer the type from the data
        inferred_array = pa.array(column_data, type=inferred_arrow_type)
    # detailed error handling should happen in fallback cases
    except (pa.ArrowInvalid, pyarrow.ArrowTypeError):
        logger.warning(
            f"Default conversion to `{inferred_arrow_type}` for `data_type={dlt_data_type}` failed."
            " Using fallback strategies."
        )

    def _first_non_none(types_: Sequence[AnyType]) -> bool:
        # determine the first non-null value to guide fallbacks
        first_non_none = None
        for _v in column_data:
            if _v is not None:
                first_non_none = _v
                break
        return isinstance(first_non_none, types_)  # type: ignore[arg-type]

    # case 1 & 2: pyarrow infers the type (e.g., float, string) THEN cast it to the dlt specified type; less constraints than the base case
    # for example, this handles when backends return decimals as floats or strings
    if inferred_array is None and dlt_data_type is not None:
        try:
            raw_array = pa.array(column_data)
            inferred_array = cast_arrow_array_as_column_schema(
                raw_array, caps, column_schema, tz, safe_arrow_conversion
            )
        except PyToArrowConversionException:
            raise
        except Exception as e:
            if (
                (dlt_data_type in ("text", "json"))
                and pa.types.is_string(inferred_arrow_type)
                and _first_non_none((list, dict))
            ):
                # this is handled by fallback case 2 for text/json with nested values
                logger.warning(
                    f"Received `data_type='{dlt_data_type}'`, data requires serialization to"
                    " string, slowing extraction. Cast the JSON field to STRING in your database"
                    " system to improve performance. For example, create and extract data from an"
                    " SQL VIEW that SELECT with CAST."
                )
            else:
                raise PyToArrowConversionException(
                    data_type=dlt_data_type,
                    inferred_arrow_type=inferred_arrow_type,
                    details=f"This conversion is currently unsupported by dlt ({str(e)})",
                ) from e

    # case 2: encode Sequence and Mapping types (list, tuples, set, dict, etc.) to JSON strings
    # This logic needs to be before case 3, otherwise pyarrow might infer the deserialized JSON object as a `pyarrow.struct` instead of `pyarrow.string`
    if inferred_array is None and dlt_data_type in (
        "json",
        "text",
    ):
        # depending on the backend, JSON columns are inferred as data_type="text"
        json_serialized_values: list[Union[bytes, None]] = []
        for value in column_data:
            if value is None:
                json_serialized_values.append(None)
                continue
            try:
                json_serialized_values.append(json.dumpb(value))
            except TypeError as e:
                raise PyToArrowConversionException(
                    data_type=dlt_data_type,
                    inferred_arrow_type=inferred_arrow_type,
                    details="dlt failed to a JSON-serializable type.",
                ) from e

        inferred_array = pa.array(json_serialized_values).cast(pa.string())

    # case 3: encode Python types unsupported by Arrow. Simple types are converted to strings and complex types to common structures (dict, list)
    # This catches specialized SQL types like `Ranges`
    if inferred_array is None and dlt_data_type is None:
        try:
            inferred_array = pa.array(column_data)
        except (pa.ArrowInvalid, pyarrow.ArrowTypeError) as e:
            logger.warning(
                f"Type can't be inferred by `pyarrow` {e.args[0]}. Values will be encoded as in a"
                " loop, slowing extraction."
            )
            encoded_values: list[Union[None, Mapping[Any, Any], Sequence[Any], str]] = []
            for value in column_data:
                if value is None:
                    encoded_values.append(None)
                    continue
                try:
                    # the 3 types match those supported by `map_nested_in_place()`
                    if isinstance(value, (tuple, dict, list)):
                        encoded_value = map_nested_values_in_place(custom_encode, value)
                    # convert set to list
                    elif isinstance(value, set):
                        encoded_value = map_nested_values_in_place(custom_encode, list(value))
                    # no nesting
                    else:
                        encoded_value = custom_encode(value)  # type: ignore[assignment]
                    encoded_values.append(encoded_value)
                except TypeError as e:
                    raise PyToArrowConversionException(
                        data_type=dlt_data_type,
                        inferred_arrow_type=inferred_arrow_type,
                        details="dlt failed to encode values to an Arrow-compatible type.",
                    ) from e

            inferred_array = pa.array(encoded_values)

    return inferred_array


def cast_arrow_array_as_column_schema(
    raw_array: pyarrow.Array,
    caps: DestinationCapabilitiesContext,
    column_schema: TColumnSchema,
    tz: str,
    safe_arrow_conversion: bool,
) -> pyarrow.Array:
    from dlt.common.libs.pyarrow import pyarrow as pa

    dlt_data_type = column_schema.get("data_type")
    inferred_arrow_type = (
        get_py_arrow_datatype(column_schema, caps, tz) if dlt_data_type is not None else None
    )
    inferred_array = None
    try:
        inferred_array = raw_array.cast(inferred_arrow_type, safe=safe_arrow_conversion)
    except (pa.ArrowInvalid, pyarrow.ArrowTypeError, pyarrow.ArrowNotImplementedError) as e:
        # TODO add specific error handling as we encounter them
        error_msg = e.args[0]
        if (
            "would cause data loss"
            in error_msg  # specific pyarrow error related to precision loss (i.e., conversion to decimal)
            and dlt_data_type == "decimal"
            and pa.types.is_decimal(inferred_arrow_type)
        ):
            # TODO provide user interface for safe_arrow_conversion=False and include in this error message
            raise PyToArrowConversionException(
                data_type=dlt_data_type,
                inferred_arrow_type=inferred_arrow_type,
                details=(
                    f"Insufficient decimal precision {error_msg}. Consider setting `precision`"
                    " and `scale` hints:"
                    " https://dlthub.com/docs/general-usage/schema/#tables-and-columns"
                ),
            ) from e
        elif dlt_data_type == "timestamp" and "Failed to parse string" in error_msg:
            if "expected a zone offset" in error_msg:
                # expected tz-aware timestamps in column_data
                actual_type = pyarrow.timestamp(inferred_arrow_type.unit, None)

            elif "expected no zone offset" in error_msg:
                # expected naive timestamps in column_data
                actual_type = pyarrow.timestamp(inferred_arrow_type.unit, tz)
            else:
                raise PyToArrowConversionException(
                    data_type=dlt_data_type,
                    inferred_arrow_type=inferred_arrow_type,
                    details=f"Timestamp conversion unsupported by dlt ({error_msg})",
                ) from e
            # cast strings to date-times
            raw_array = raw_array.cast(actual_type)
            _, inferred_array = normalize_py_arrow_item_column(
                column_schema, actual_type, raw_array
            )
        elif dlt_data_type == "time" and "function cast_time" in error_msg:
            if "from string to" in error_msg:
                n = len(raw_array)
                is_null = None
                is_null_b = None
                null_count = 0

                data_b = pyarrow.allocate_buffer(n * 8, resizable=False)
                data = memoryview(data_b).cast("q")

                def allocate_lazy_null_mask() -> None:
                    nonlocal is_null, data, is_null_b
                    if is_null is None:
                        nbytes = (n + 7) // 8
                        is_null_b = pa.allocate_buffer(nbytes, resizable=False)
                        ctypes.memset(is_null_b.address, 0xFF, nbytes)  # start all-valid
                        is_null = memoryview(is_null_b).cast("B")

                for i, scalar in enumerate(raw_array):
                    if not scalar.is_valid:
                        # mark null in bitmap
                        allocate_lazy_null_mask()
                        byte = i >> 3
                        bit = i & 7
                        is_null[byte] &= ~(1 << bit)
                        null_count += 1
                        continue

                    s = scalar.as_py()  # Python str
                    t = time.fromisoformat(s)  # HH:MM | HH:MM:SS | HH:MM:SS.ffffff

                    data[i] = (
                        (t.hour * 3600 + t.minute * 60 + t.second) * 1_000_000
                    ) + t.microsecond

                # Zero-copy build of Arrow int64 with mask; then logical view to time64[us]
                arr_i64 = pyarrow.Array.from_buffers(
                    pa.int64(), n, buffers=[is_null_b, data_b], null_count=null_count
                )  # data zero-copies from NumPy
                inferred_array = arr_i64.view(pa.time64("us"))  # zero-copy reinterpret
                # ts  = pyarrow.compute.strptime(raw_array, format="%H:%M:%S", unit="us")
                # inferred_array = pyarrow.compute.cast(ts, pa.time64("us"))

            elif "from duration" in error_msg:
                # duration and time (if wrapping is not needed) have the same representation
                inferred_array = raw_array.view(pa.time64(raw_array.type.unit))

            else:
                raise PyToArrowConversionException(
                    data_type=dlt_data_type,
                    inferred_arrow_type=inferred_arrow_type,
                    details=f"Time conversion unsupported by dlt ({error_msg})",
                ) from e
        # fallback: nested arrow -> json strings when casting to text/json
        elif (
            pa.types.is_nested(raw_array.type)
            and pa.types.is_string(inferred_arrow_type)
            and dlt_data_type in ("text", "json")
        ):
            # log a warning similar to numpy case 2, explaining serialization and destination caps
            logger.warning(
                "Received `data_type='%s'`, Arrow nested values require serialization to STRING."
                " They will be JSON-encoded, which may slow extraction. If your destination"
                " supports nested types, enable them (`supports_nested_types=True`) and provide a"
                " 'json' column with 'x-nested-type' hint to load nested values natively."
                " Otherwise, consider casting to STRING upstream (e.g., in a VIEW).",
                dlt_data_type,
            )
            try:
                # convert nested arrow values to python and json-encode. this mirrors the generic
                # json/text fallback (case 2) used for numpy inputs, but works on arrow arrays.
                py_values = raw_array.to_pylist()
                json_bytes = [None if v is None else json.dumpb(v) for v in py_values]
                inferred_array = pa.array(json_bytes).cast(pa.string())
            except TypeError as enc_err:
                raise PyToArrowConversionException(
                    data_type=dlt_data_type,
                    inferred_arrow_type=inferred_arrow_type,
                    details="dlt failed to encode nested values as JSON strings.",
                ) from enc_err
        else:
            raise PyToArrowConversionException(
                data_type=dlt_data_type,
                inferred_arrow_type=inferred_arrow_type,
                details=f"This conversion is currently unsupported by dlt ({error_msg})",
            ) from e
    assert inferred_array
    return inferred_array


def cast_arrow_as_columns_schema(
    item: TAnyArrowItem,
    columns: TTableSchemaColumns,
    caps: DestinationCapabilitiesContext,
    tz: str,
    safe_arrow_conversion: bool = True,
) -> TAnyArrowItem:
    """Cast an Arrow item to match dlt column schema.

    Performs per-column casts using Arrow, with fallbacks equivalent to
    `cast_arrow_array_as_column_schema`. Timestamp timezone normalization is
    applied when required by the dlt schema. The result preserves the input
    item's schema metadata and per-field metadata.

    Args:
        item: Arrow table or record batch to cast.
        columns: dlt table schema columns containing data type hints.
        caps: capabilities of the storage backend.
        tz: time zone identifier used to resolve timestamp hints.
        safe_arrow_conversion: if False, truncation and loss of precision is allowed.

    Returns:
        The same kind of Arrow object (Table or RecordBatch) with columns cast to
        the requested dlt types. Schema and field metadata are preserved.
    """
    from dlt.common.libs.pyarrow import pyarrow as pa

    schema = item.schema
    arrays_out = []
    fields_out = []

    for idx, fld in enumerate(schema):
        name = fld.name
        col_schema = columns[name]
        col_data = item.column(idx)

        # default: keep original
        out_array = col_data
        out_nullable = fld.nullable
        out_type = fld.type

        out_nullable = col_schema.get("nullable", fld.nullable)

        # final cast to the exact target type using fallbacks
        try:
            cast_arr = cast_arrow_array_as_column_schema(
                out_array, caps, col_schema, tz, safe_arrow_conversion
            )
        except PyToArrowConversionException as e:
            e.field_name = name
            raise

        out_array = cast_arr
        out_type = cast_arr.type

        # build new field preserving metadata
        new_field = pa.field(name, out_type, nullable=out_nullable, metadata=fld.metadata)
        fields_out.append(new_field)
        arrays_out.append(out_array)

    # preserve schema-level metadata
    new_schema = pa.schema(fields_out, metadata=schema.metadata)

    return item.__class__.from_arrays(arrays_out, schema=new_schema)


def row_tuples_to_arrow(
    rows: TDataItems,
    caps: DestinationCapabilitiesContext,
    columns: TTableSchemaColumns,
    tz: str,
    safe_arrow_conversion: bool = True,
) -> Any:  # pyarrow.Table
    """Converts the rows to an arrow table using the columns schema.
    1. Pivot rows into columns.
    2. Convert columns to pyarrow arrays; coerce type and nullability; exclude types unsupported by arrow
    3. Create table
    4. Remove columns full of null values

    Args:
        rows: data items
        caps: capabilities of the storage backend
        columns: dlt hints about the table columns (e.g., data type, nullabe)
        tz: time zone identifier
        safe_arrow_conversion: if False, truncation and loss of precision is allowed
            ref: https://arrow.apache.org/docs/python/generated/pyarrow.compute.CastOptions.html#pyarrow.compute.CastOptions

    Returns:
        an arrow Table
    """
    from dlt.common.libs.pyarrow import pyarrow as pa

    columnar = transpose_rows_to_columns(rows, column_names=columns.keys())

    arrow_arrays = []
    arrow_fields = []
    for column_name, column_data in columnar.items():
        column_schema = columns[column_name]

        try:
            arrow_array = convert_numpy_to_arrow(
                column_data, caps, column_schema, tz, safe_arrow_conversion
            )
        # TODO if converting to arrow fail, should we raise or skip column?
        except PyToArrowConversionException as e:
            e.field_name = column_name
            raise

        field = pa.field(
            name=column_name, type=arrow_array.type, nullable=column_schema.get("nullable", True)
        )
        arrow_arrays.append(arrow_array)
        arrow_fields.append(field)

    # NOTE careful when casting, modifying types, or enforcing schemas in place. Arrow issues are common
    # This can corrupt the data when writing to Parquet
    # ref: https://github.com/apache/arrow/issues/43146
    # ref: https://github.com/apache/arrow/issues/41667
    arrow_table = pa.Table.from_arrays(arrow_arrays, schema=pa.schema(arrow_fields))
    return arrow_table


class NameNormalizationCollision(ValueError):
    def __init__(self, reason: str) -> None:
        msg = f"Arrow column name collision after input data normalization. {reason}"
        super().__init__(msg)


def add_arrow_metadata(
    item: Union[pyarrow.Table, pyarrow.RecordBatch], metadata: dict[str, Any]
) -> pyarrow.Table:
    # Get current metadata or initialize empty
    schema = item.schema
    current = schema.metadata or {}

    # Convert new metadata to bytes and merge
    update = {k.encode("utf-8"): v.encode("utf-8") for k, v in metadata.items()}
    merged = current.copy()
    merged.update(update)

    # Apply updated schema
    new_schema = schema.with_metadata(merged)

    # Rebuild the object with updated schema
    if isinstance(item, pyarrow.Table):
        return pyarrow.Table.from_arrays(item.columns, schema=new_schema)
    else:  # RecordBatch
        return pyarrow.RecordBatch.from_arrays(item.columns, schema=new_schema)


def set_plus0000_timezone_to_utc(tbl: pyarrow.Table) -> pyarrow.Table:
    """
    Convert any +00:00 timestamp columns to UTC.
    Returns the original table object if nothing needed fixing.
    """
    arrays, fields = [], []
    changed = False

    for col, fld in zip(tbl.columns, tbl.schema):
        if pyarrow.types.is_timestamp(fld.type) and fld.type.tz == "+00:00":
            changed = True
            new_type = pyarrow.timestamp(fld.type.unit, "UTC")
            arrays.append(pyarrow.compute.cast(col, new_type))
            fields.append(pyarrow.field(fld.name, new_type, fld.nullable, fld.metadata))
        else:
            arrays.append(col)
            fields.append(fld)

    if not changed:
        return tbl

    new_schema = pyarrow.schema(fields, metadata=tbl.schema.metadata)
    return pyarrow.Table.from_arrays(arrays, schema=new_schema)


def cast_date64_columns_to_timestamp(tbl: pyarrow.Table, tz: Optional[str] = None) -> pyarrow.Table:
    """
    Cast any date64 columns to timestamp with microsecond precision, preserving the
    semantic time values. Uses pyarrow.compute.cast on the column (works for chunked arrays)
    to cast from milliseconds (date64) to microseconds (timestamp[us]).

    Args:
        tbl: Input Arrow table.
        tz: Optional timezone to annotate the resulting timestamp with (e.g. "UTC").
            If None (default), produces a naive timestamp.

    Returns:
        A new table with date64 columns cast to timestamp[us] (optionally tz-aware),
        or the original table if no date64 columns were found.
    """
    arrays, fields = [], []
    changed = False

    for col, fld in zip(tbl.columns, tbl.schema):
        if pyarrow.types.is_date64(fld.type):
            changed = True
            unit = "us"
            new_type = pyarrow.timestamp(unit, tz)
            # Rescale from ms (date64) to us (timestamp).
            arrays.append(pyarrow.compute.cast(col, new_type))
            fields.append(pyarrow.field(fld.name, new_type, fld.nullable, fld.metadata))
        else:
            arrays.append(col)
            fields.append(fld)

    if not changed:
        return tbl

    new_schema = pyarrow.schema(fields, metadata=tbl.schema.metadata)
    return pyarrow.Table.from_arrays(arrays, schema=new_schema)



from .specs.base_configuration import (
    configspec,
    is_valid_hint,
    is_secret_hint,
    resolve_type,
    NotResolved,
)
from .specs import known_sections
from .resolve import resolve_configuration, inject_section
from .inject import with_config, last_config, get_fun_spec, create_resolved_partial

from .exceptions import (
    ConfigFieldMissingException,
    ConfigValueCannotBeCoercedException,
    ConfigFileNotFoundException,
    ConfigurationValueError,
)


__all__ = [
    "configspec",
    "is_valid_hint",
    "is_secret_hint",
    "NotResolved",
    "resolve_type",
    "known_sections",
    "resolve_configuration",
    "inject_section",
    "with_config",
    "last_config",
    "get_fun_spec",
    "ConfigFieldMissingException",
    "ConfigValueCannotBeCoercedException",
    "ConfigFileNotFoundException",
    "ConfigurationValueError",
]



from typing import Any, NamedTuple, Tuple, Iterable, Mapping
import tomlkit
from tomlkit.items import Table as TOMLTable
from tomlkit.container import Container as TOMLContainer

from dlt.common.configuration.specs.base_configuration import is_hint_not_resolvable
from dlt.common.configuration.const import TYPE_EXAMPLES
from dlt.common.configuration.specs import (
    BaseConfiguration,
    is_base_configuration_inner_hint,
    extract_inner_hint,
)
from dlt.common.data_types import py_type_to_sc_type
from dlt.common.typing import AnyType, is_optional_type, is_subclass


class WritableConfigValue(NamedTuple):
    name: Any
    hint: AnyType
    default_value: Any
    sections: Tuple[str, ...]


def generate_typed_example(name: str, hint: AnyType) -> Any:
    inner_hint = extract_inner_hint(hint)
    try:
        sc_type = py_type_to_sc_type(inner_hint)
        if sc_type in TYPE_EXAMPLES.keys():
            return TYPE_EXAMPLES[sc_type]
        if sc_type == "bigint":
            return 0
        if sc_type == "double":
            return 1.0
        if sc_type == "bool":
            return True
        if sc_type == "json":
            return "json"
        if sc_type in ("wei", "decimal"):
            return "1.0"
        raise TypeError(sc_type)
    except TypeError:
        return name


def write_value(
    toml_table: TOMLTable,
    name: str,
    hint: AnyType,
    overwrite_existing: bool,
    default_value: Any = None,
    is_default_of_interest: bool = False,
) -> None:
    # skip if table contains the name already
    if name in toml_table and not overwrite_existing:
        return
    # do not dump nor resolvable and optional fields if they are not of special interest
    if (
        is_hint_not_resolvable(hint) or is_optional_type(hint) or default_value is not None
    ) and not is_default_of_interest:
        return
    # get the inner hint to generate cool examples
    hint = extract_inner_hint(hint)
    if is_base_configuration_inner_hint(hint):
        inner_table = tomlkit.table(is_super_table=True)
        write_spec(inner_table, hint(), default_value, overwrite_existing)
        if len(inner_table) > 0:
            toml_table[name] = inner_table
    else:
        if default_value is None:
            example_value = generate_typed_example(name, hint)
            # complex types do not get default values for now (#2531)
            if example_value == "json":
                return
            toml_table[name] = example_value
            # tomlkit not supporting comments on boolean
            if not isinstance(example_value, bool):
                toml_table[name].comment("fill this in!")
        else:
            toml_table[name] = default_value


def write_spec(
    toml_table: TOMLTable,
    config: BaseConfiguration,
    initial_value: Mapping[str, Any],
    overwrite_existing: bool,
) -> None:
    for name, hint in config.get_resolvable_fields().items():
        # use initial value
        initial_ = initial_value.get(name) if initial_value else None
        # use default value stored in config
        default_value = getattr(config, name, None)

        # check if field is of particular interest and should be included if it has default
        is_default_of_interest = name in config.__config_gen_annotations__

        # if initial is different from default, it is of interest as well
        if initial_ is not None:
            is_default_of_interest = is_default_of_interest or (initial_ != default_value)

        write_value(
            toml_table,
            name,
            hint,
            overwrite_existing,
            default_value=initial_ or default_value,
            is_default_of_interest=is_default_of_interest,
        )


def write_values(
    toml: TOMLContainer, values: Iterable[WritableConfigValue], overwrite_existing: bool
) -> None:
    # TODO: decouple writers from a particular object model ie. TOML
    for value in values:
        toml_table: TOMLTable = toml  # type: ignore
        for section in value.sections:
            if section not in toml_table:
                inner_table = tomlkit.table(is_super_table=True)
                toml_table[section] = inner_table
                toml_table = inner_table
            else:
                toml_table = toml_table[section]  # type: ignore

        write_value(
            toml_table,
            value.name,
            value.hint,
            overwrite_existing,
            default_value=value.default_value,
            is_default_of_interest=True,
        )



"""Implements SupportsTracking"""
import contextlib
from typing import Any, Union, Dict, Optional
import humanize

from dlt.common import logger
from dlt.common.pendulum import pendulum
from dlt.common.utils import digest128
from dlt.common.runtime.exec_info import github_info
from dlt.common.runtime.anon_tracker import track as dlthub_telemetry_track
from dlt.common.runtime.slack import send_slack_message
from dlt.common.pipeline import LoadInfo, ExtractInfo, SupportsPipeline

from dlt.pipeline.typing import TPipelineStep
from dlt.pipeline.trace import PipelineTrace, PipelineStepTrace

try:
    from sentry_sdk import Scope
    from sentry_sdk.tracing import Span, Transaction

    def _add_sentry_tags(span: Span, pipeline: SupportsPipeline) -> None:
        span.set_tag("pipeline_name", pipeline.pipeline_name)
        if pipeline.destination:
            span.set_tag("destination", pipeline.destination.destination_name)
        if pipeline.dataset_name:
            span.set_tag("dataset_name", pipeline.dataset_name)

except ImportError:
    # sentry is optional dependency and enabled only when RuntimeConfiguration.sentry_dsn is set
    pass


def slack_notify_load_success(incoming_hook: str, load_info: LoadInfo, trace: PipelineTrace) -> int:
    """Sends a markdown formatted success message and returns http status code from the Slack incoming hook"""
    try:
        author = github_info().get("github_user", "")
        if author:
            author = f":hard-hat:{author}'s "

        total_elapsed = pendulum.now().diff(trace.started_at)

        def _get_step_elapsed(step: PipelineStepTrace) -> str:
            if not step:
                return ""
            elapsed = step.finished_at - step.started_at
            return f"`{step.step.upper()}`: _{humanize.precisedelta(elapsed)}_ "

        load_step = trace.steps[-1]
        normalize_step = next((step for step in trace.steps if step.step == "normalize"), None)
        extract_step = next((step for step in trace.steps if step.step == "extract"), None)

        message = f"""The {author}pipeline *{load_info.pipeline.pipeline_name}* just loaded *{len(load_info.loads_ids)}* load package(s) to destination *{load_info.destination_type}* and into dataset *{load_info.dataset_name}*.
🚀 *{humanize.precisedelta(total_elapsed)}* of which {_get_step_elapsed(load_step)}{_get_step_elapsed(normalize_step)}{_get_step_elapsed(extract_step)}"""

        send_slack_message(incoming_hook, message)
        return 200
    except Exception as ex:
        logger.warning(f"Slack notification could not be sent: {str(ex)}")
        return -1


def on_start_trace(trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline) -> None:
    if pipeline.run_context.runtime_config.sentry_dsn:
        # print(f"START SENTRY TX: {trace.transaction_id} SCOPE: {Hub.current.scope}"
        transaction = Scope.get_current_scope().start_transaction(name=step, op=step)
        if isinstance(transaction, Transaction):
            _add_sentry_tags(transaction, pipeline)
            transaction.__enter__()


def on_start_trace_step(
    trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
) -> None:
    if pipeline.run_context.runtime_config.sentry_dsn:
        # print(f"START SENTRY SPAN {trace.transaction_id}:{trace_step.span_id} SCOPE: {Hub.current.scope}")
        span = Scope.get_current_scope().start_span(description=step, op=step)
        _add_sentry_tags(span, pipeline)
        span.__enter__()


def _build_base_props(
    pipeline: SupportsPipeline,
) -> Dict[str, Union[str, None]]:
    return {
        "destination_name": pipeline.destination.destination_name if pipeline.destination else None,
        "destination_type": pipeline.destination.destination_type if pipeline.destination else None,
        "pipeline_name_hash": digest128(pipeline.pipeline_name),
        "dataset_name_hash": digest128(pipeline.dataset_name) if pipeline.dataset_name else None,
        "default_schema_name_hash": (
            digest128(pipeline.default_schema_name) if pipeline.default_schema_name else None
        ),
    }


def on_first_dataset_access(
    pipeline: SupportsPipeline,
    success: bool,
    schema_name: Optional[str] = None,
) -> None:
    """Track the first dataset access telemetry event for the given pipeline instance."""
    props = {
        "success": success,
        **_build_base_props(pipeline=pipeline),
        # The schema may differ from the default pipeline schema
        "requested_schema_name_hash": digest128(schema_name) if schema_name else None,
    }
    dlthub_telemetry_track("pipeline", "access_dataset", props)


def on_end_trace_step(
    trace: PipelineTrace,
    step: PipelineStepTrace,
    pipeline: SupportsPipeline,
    step_info: Any,
    send_state: bool,
) -> None:
    if pipeline.run_context.runtime_config.sentry_dsn:
        Scope.get_current_scope().span.__exit__(None, None, None)
    # disable automatic slack messaging until we can configure messages themselves
    # if step.step == "load":
    #     if pipeline.runtime_config.slack_incoming_hook and step.step_exception is None:
    #         slack_notify_load_success(pipeline.runtime_config.slack_incoming_hook, step_info, trace)
    props = {
        "elapsed": (step.finished_at - trace.started_at).total_seconds(),
        "success": step.step_exception is None,
        **_build_base_props(pipeline=pipeline),
        "transaction_id": trace.transaction_id,
    }
    if step.step == "extract" and step_info:
        assert isinstance(step_info, ExtractInfo)
        props["extract_data"] = step_info.extract_data_info
    if step.step == "load" and step_info:
        assert isinstance(step_info, LoadInfo)
        props["destination_fingerprint"] = step_info.destination_fingerprint
    dlthub_telemetry_track("pipeline", step.step, props)


def on_end_trace(trace: PipelineTrace, pipeline: SupportsPipeline, send_state: bool) -> None:
    if pipeline.run_context.runtime_config.sentry_dsn:
        # print(f"---END SENTRY TX: {trace.transaction_id} SCOPE: {Hub.current.scope}")
        Scope.get_current_scope().transaction.__exit__(None, None, None)



from copy import copy
from typing import List, Dict, Sequence, Set, Any, Optional, Tuple
from functools import lru_cache

from dlt.common.data_types.typing import TDataType
from dlt.common.destination.capabilities import adjust_column_schema_to_capabilities
from dlt.common import logger
from dlt.common.json import json
from dlt.common.json import custom_pua_decode, may_have_pua
from dlt.common.schema import utils
from dlt.common.schema.typing import (
    TColumnSchema,
    TPartialTableSchema,
    TSchemaEvolutionMode,
    TTableSchemaColumns,
    TSchemaContractDict,
)
from dlt.common.schema.utils import has_table_seen_data, is_complete_column
from dlt.common.schema.exceptions import CannotCoerceColumnException, CannotCoerceNullException
from dlt.common.schema import TSchemaUpdate, Schema
from dlt.common.storages.load_storage import LoadStorage
from dlt.common.storages import NormalizeStorage
from dlt.common.storages.data_item_storage import DataItemStorage
from dlt.common.time import normalize_timezone
from dlt.common.typing import VARIANT_FIELD_FORMAT, DictStrAny, REPattern, StrAny, TDataItem

from dlt.normalize.configuration import NormalizeConfiguration
from dlt.normalize.items_normalizers.base import ItemsNormalizer


class JsonLItemsNormalizer(ItemsNormalizer):
    def __init__(
        self,
        item_storage: DataItemStorage,
        load_storage: LoadStorage,
        normalize_storage: NormalizeStorage,
        schema: Schema,
        load_id: str,
        config: NormalizeConfiguration,
    ) -> None:
        super().__init__(item_storage, load_storage, normalize_storage, schema, load_id, config)
        self._table_contracts: Dict[str, TSchemaContractDict] = {}
        self._filtered_tables: Set[str] = set()
        self._filtered_tables_columns: Dict[str, Dict[str, TSchemaEvolutionMode]] = {}
        # quick access to column schema for writers below
        self._column_schemas: Dict[str, TTableSchemaColumns] = {}
        self._null_only_columns: Dict[str, Set[str]] = {}
        self._shorten_fragments = lru_cache(maxsize=None)(self.schema.naming.shorten_fragments)
        # cache coercion interface from data item normalizer
        _din = self.schema.data_item_normalizer
        self._type_map = _din.py_type_to_sc_type_map
        self._can_coerce_type = _din.can_coerce_type
        self._coerce_type = _din.coerce_type
        self._py_type_to_sc_type = _din.py_type_to_sc_type

    @property
    def null_only_columns(self) -> Dict[str, Set[str]]:
        return self._null_only_columns

    def _filter_columns(
        self, filtered_columns: Dict[str, TSchemaEvolutionMode], row: DictStrAny
    ) -> DictStrAny:
        for name, mode in filtered_columns.items():
            if name in row:
                if mode == "discard_row":
                    return None
                elif mode == "discard_value":
                    row.pop(name)
        return row

    def _normalize_chunk(
        self, root_table_name: str, items: List[TDataItem], may_have_pua: bool, skip_write: bool
    ) -> TSchemaUpdate:
        column_schemas = self._column_schemas
        schema_update: TSchemaUpdate = {}
        schema = self.schema
        schema_name = schema.name
        normalize_data_fun = self.schema.normalize_data_item

        for item in items:
            items_gen = normalize_data_fun(item, self.load_id, root_table_name)
            try:
                should_descend: bool = None
                # use send to prevent descending into child rows when row was discarded
                while row_info := items_gen.send(should_descend):
                    should_descend = True
                    (table_name, parent_path, ident_path), row = row_info
                    parent_table = self._shorten_fragments(*parent_path)

                    # rows belonging to filtered out tables are skipped
                    if table_name in self._filtered_tables:
                        # stop descending into further rows
                        should_descend = False
                        continue

                    # filter row, may eliminate some or all fields
                    row = self._filter_row(table_name, row)
                    # do not process empty rows
                    if not row:
                        should_descend = False
                        continue

                    # filter columns or full rows if schema contract said so
                    # do it before schema inference in `coerce_row` to not trigger costly migration code
                    filtered_columns = self._filtered_tables_columns.get(table_name, None)
                    if filtered_columns:
                        row = self._filter_columns(filtered_columns, row)  # type: ignore[arg-type]
                        # if whole row got dropped
                        if not row:
                            should_descend = False
                            continue

                    # decode pua types
                    if may_have_pua:
                        for k, v in row.items():
                            row[k] = custom_pua_decode(v)  # type: ignore

                    # coerce row of values into schema table, generating partial table with new columns if any
                    row, partial_table = self._coerce_row(table_name, parent_table, row)

                    # if we detect a migration, check schema contract
                    if partial_table:
                        schema_contract = self._table_contracts.setdefault(
                            table_name,
                            schema.resolve_contract_settings_for_table(
                                table_name
                                if table_name in schema.tables
                                else parent_table or table_name
                            ),  # parent_table, if present, exists in the schema
                        )

                        partial_table, filters = schema.apply_schema_contract(
                            schema_contract, partial_table, data_item=row
                        )
                        if filters:
                            for entity, name, mode in filters:
                                if entity == "tables":
                                    self._filtered_tables.add(name)
                                elif entity == "columns":
                                    filtered_columns = self._filtered_tables_columns.setdefault(
                                        table_name, {}
                                    )
                                    filtered_columns[name] = mode

                        if partial_table is None:
                            # discard migration and row
                            should_descend = False
                            continue
                        # theres a new table or new columns in existing table
                        # update schema and save the change
                        schema.update_table(partial_table, normalize_identifiers=False)
                        table_updates = schema_update.setdefault(table_name, [])
                        table_updates.append(partial_table)

                        # update our columns
                        column_schemas[table_name] = schema.get_table_columns(table_name)

                        # apply new filters
                        if filtered_columns and filters:
                            row = self._filter_columns(filtered_columns, row)
                            # do not continue if new filters skipped the full row
                            if not row:
                                should_descend = False
                                continue

                    # get current columns schema
                    columns = column_schemas.get(table_name)
                    if not columns:
                        columns = schema.get_table_columns(table_name)
                        column_schemas[table_name] = columns
                    # store row
                    # TODO: store all rows for particular items all together after item is fully completed
                    #   will be useful if we implement bad data sending to a table
                    # we skip write when discovering schema for empty file
                    if not skip_write:
                        self.item_storage.write_data_item(
                            self.load_id, schema_name, table_name, row, columns
                        )
            except StopIteration:
                pass

        return schema_update

    def _coerce_row(
        self, table_name: str, parent_table: str, row: StrAny
    ) -> Tuple[DictStrAny, TPartialTableSchema]:
        """Fits values of fields present in `row` into a schema of `table_name`. Will coerce values into data types and infer new tables and column schemas.

        Method expects that field names in row are already normalized.
        * if table schema for `table_name` does not exist, new table is created
        * if column schema for a field in `row` does not exist, it is inferred from data
        * if incomplete column schema (no data type) exists, column is inferred from data and existing hints are applied
        * fields with None value are removed

        Returns tuple with row with coerced values and a partial table containing just the newly added columns or None if no changes were detected
        """
        # get existing or create a new table
        updated_table_partial: TPartialTableSchema = None
        table = self.schema._schema_tables.get(table_name)
        if not table:
            table = utils.new_table(table_name, parent_table)
        table_columns = table["columns"]

        new_row: DictStrAny = {}
        type_map = self._type_map
        can_coerce_type = self._can_coerce_type
        coerce_type = self._coerce_type
        for col_name, v in row.items():
            # skip None values, we should infer the types later
            if v is None:
                # just check if column is nullable if it exists
                new_col_def = self._coerce_null_value(table_columns, table_name, col_name)
                new_col_name = col_name
            else:
                # fast path: known column with matching type
                if existing_column := table_columns.get(col_name):
                    if col_type := existing_column.get("data_type"):
                        py_type = type_map.get(type(v))
                        if py_type == col_type:
                            # happy path: complete column, type matches, no coercion needed
                            new_row[col_name] = v
                            continue
                        if py_type and can_coerce_type(col_type, py_type):
                            # type mismatch but coercion is known - try fast path
                            try:
                                new_v = coerce_type(col_type, py_type, v)
                                if col_type == "timestamp":
                                    timezone = existing_column.get("timezone", True)
                                    new_v = normalize_timezone(new_v, timezone)
                                new_row[col_name] = new_v
                                continue
                            except (ValueError, SyntaxError):
                                # coercion failed - fall through to slow path for variant handling
                                pass
                # new column, incomplete column or variant
                new_col_name, new_col_def, new_v = self._coerce_non_null_value(
                    table_columns, table_name, col_name, v
                )
                new_row[new_col_name] = new_v
            if new_col_def:
                if not updated_table_partial:
                    # create partial table with only the new columns
                    updated_table_partial = copy(table)
                    updated_table_partial["columns"] = {}
                updated_table_partial["columns"][new_col_name] = new_col_def

        return new_row, updated_table_partial

    def _infer_column(
        self,
        k: str,
        v: Any,
        data_type: TDataType = None,
        is_variant: bool = False,
        table_name: str = None,
    ) -> Optional[TColumnSchema]:
        if v is None and data_type is None:
            if self.schema._infer_hint("not_null", k):
                raise CannotCoerceNullException(self.schema.name, table_name, k)
            return None
        else:
            column_schema = TColumnSchema(
                name=k,
                data_type=data_type or self._infer_column_type(v, k),
                nullable=not self.schema._infer_hint("not_null", k),
            )
        # check other preferred hints that are available
        for hint in self.schema._compiled_hints:
            # already processed
            if hint == "not_null":
                continue
            column_prop = utils.hint_to_column_prop(hint)
            hint_value = self.schema._infer_hint(hint, k)
            # set only non-default values
            if not utils.has_default_column_prop_value(column_prop, hint_value):
                column_schema[column_prop] = hint_value

        if is_variant:
            column_schema["variant"] = is_variant
        return column_schema

    def _coerce_null_value(
        self, table_columns: TTableSchemaColumns, table_name: str, col_name: str
    ) -> Optional[TColumnSchema]:
        """Raises on not-nullable columns, tracks null-only column names"""
        existing_column = table_columns.get(col_name)
        if existing_column and utils.is_complete_column(existing_column):
            if not utils.is_nullable_column(existing_column):
                raise CannotCoerceNullException(self.schema.name, table_name, col_name)
        else:
            if not existing_column and self.schema._infer_hint("not_null", col_name):
                raise CannotCoerceNullException(self.schema.name, table_name, col_name)
            self._null_only_columns.setdefault(table_name, set()).add(col_name)
        return None

    def _coerce_non_null_value(
        self,
        table_columns: TTableSchemaColumns,
        table_name: str,
        col_name: str,
        v: Any,
        is_variant: bool = False,
    ) -> Tuple[str, TColumnSchema, Any]:
        new_column: TColumnSchema = None
        existing_column = table_columns.get(col_name)
        # if column exist but is incomplete then keep it as new column
        if existing_column and not utils.is_complete_column(existing_column):
            new_column = existing_column
            existing_column = None

        # infer type or get it from existing table
        col_type = (
            existing_column["data_type"]
            if existing_column
            else self._infer_column_type(v, col_name, skip_preferred=is_variant)
        )
        # get data type of value
        py_type = self._py_type_to_sc_type(type(v))
        # and coerce type if inference changed the python type
        try:
            coerced_v = self._coerce_type(col_type, py_type, v)
        except (ValueError, SyntaxError):
            if is_variant:
                # this is final call: we cannot generate any more auto-variants
                raise CannotCoerceColumnException(
                    self.schema.name,
                    table_name,
                    col_name,
                    py_type,
                    table_columns[col_name]["data_type"],
                    v,
                )
            # otherwise we must create variant extension to the table
            # backward compatibility for complex types: if such column exists then use it
            variant_col_name = self.naming.shorten_fragments(
                col_name, VARIANT_FIELD_FORMAT % py_type
            )
            if py_type == "json":
                old_complex_col_name = self.naming.shorten_fragments(
                    col_name, VARIANT_FIELD_FORMAT % "complex"
                )
                if old_column := table_columns.get(old_complex_col_name):
                    if old_column.get("variant"):
                        variant_col_name = old_complex_col_name
            # pass final=True so no more auto-variants can be created recursively
            return self._coerce_non_null_value(
                table_columns, table_name, variant_col_name, v, is_variant=True
            )

        # if coerced value is variant, then extract variant value
        # note: checking runtime protocols with isinstance(coerced_v, SupportsVariant): is extremely slow so we check if callable as every variant is callable
        if callable(coerced_v):  # and isinstance(coerced_v, SupportsVariant):
            coerced_v = coerced_v()
            if isinstance(coerced_v, tuple):
                # variant recovered so call recursively with variant column name and variant value
                variant_col_name = self.naming.shorten_fragments(
                    col_name, VARIANT_FIELD_FORMAT % coerced_v[0]
                )
                return self._coerce_non_null_value(
                    table_columns, table_name, variant_col_name, coerced_v[1], is_variant=True
                )

        if not existing_column:
            # adjust to current capabilities, mostly removes hints that are default for a given
            # destination
            inferred_column = adjust_column_schema_to_capabilities(
                self._infer_column(col_name, v, data_type=col_type, is_variant=is_variant),
                self.config.destination_capabilities,
            )
            # if there's incomplete new_column then merge it with inferred column
            if new_column:
                # use all values present in incomplete column to override inferred column - also the defaults
                new_column = utils.merge_column(inferred_column, new_column)
            else:
                new_column = inferred_column

        # apply timestamp when column schema is known
        if col_type == "timestamp":
            timezone = (existing_column or new_column).get("timezone", True)
            coerced_v = normalize_timezone(coerced_v, timezone)

        return col_name, new_column, coerced_v

    def _infer_column_type(self, v: Any, col_name: str, skip_preferred: bool = False) -> TDataType:
        tv = type(v)
        # try to autodetect data type
        mapped_type = utils.autodetect_sc_type(self.schema._type_detections, tv, v)
        # if not try standard type mapping
        if mapped_type is None:
            mapped_type = self._py_type_to_sc_type(tv)
        # get preferred type based on column name
        preferred_type: TDataType = None
        if not skip_preferred:
            preferred_type = self.schema.get_preferred_type(col_name)
        return preferred_type or mapped_type

    def _filter_row(self, table_name: str, row: StrAny) -> StrAny:
        # TODO: remove this. move to extract stage
        # exclude row elements according to the rules in `filter` elements of the table
        # include rules have precedence and are used to make exceptions to exclude rules
        # the procedure will apply rules from the table_name and it's all parent tables up until root
        # parent tables are computed by `normalize_break_path` function so they do not need to exist in the schema
        # note: the above is not very clean. the `parent` element of each table should be used but as the rules
        #  are typically used to prevent not only table fields but whole tables from being created it is not possible

        if not (compiled_excludes := self.schema._compiled_excludes):
            # if there are no excludes in the whole schema, no modification to a row can be made
            # most of the schema do not use them
            return row

        def _exclude(
            path: str, excludes: Sequence[REPattern], includes: Sequence[REPattern]
        ) -> bool:
            is_included = False
            is_excluded = any(exclude.search(path) for exclude in excludes)
            if is_excluded:
                # we may have exception if explicitly included
                is_included = any(include.search(path) for include in includes)
            return is_excluded and not is_included

        # break table name in components
        branch = self.naming.break_path(table_name)
        compiled_includes = self.schema._compiled_includes

        # check if any of the rows is excluded by rules in any of the tables
        for i in range(len(branch), 0, -1):  # stop is exclusive in `range`
            # start at the top level table
            c_t = self.naming.make_path(*branch[:i])
            excludes = compiled_excludes.get(c_t)
            # only if there's possibility to exclude, continue
            if excludes:
                includes = compiled_includes.get(c_t) or []
                for field_name in list(row.keys()):
                    path = self.naming.make_path(*branch[i:], field_name)
                    if _exclude(path, excludes, includes):
                        # TODO: copy to new instance
                        del row[field_name]  # type: ignore
            # if row is empty, do not process further
            if not row:
                break
        return row

    def __call__(
        self,
        extracted_items_file: str,
        root_table_name: str,
    ) -> List[TSchemaUpdate]:
        self._maybe_cancel()
        schema_updates: List[TSchemaUpdate] = []
        with self.normalize_storage.extracted_packages.storage.open_file(
            extracted_items_file, "rb"
        ) as f:
            # enumerate jsonl file line by line
            line: bytes = None
            for line_no, line in enumerate(f):
                self._maybe_cancel()
                items: List[TDataItem] = json.loadb(line)
                partial_update = self._normalize_chunk(
                    root_table_name, items, may_have_pua(line), skip_write=False
                )
                schema_updates.append(partial_update)
                logger.debug(f"Processed {line_no+1} lines from file {extracted_items_file}")
            # empty json files are when replace write disposition is used in order to truncate table(s)
            if line is None and root_table_name in self.schema.tables:
                root_table = self.schema.tables[root_table_name]
                if not has_table_seen_data(root_table):
                    # if this is a new table, add normalizer columns
                    partial_update = self._normalize_chunk(
                        root_table_name, [{}], False, skip_write=True
                    )
                    schema_updates.append(partial_update)
                self.item_storage.write_empty_items_file(
                    self.load_id,
                    self.schema.name,
                    root_table_name,
                    self.schema.get_table_columns(root_table_name),
                )
                logger.debug(
                    f"No lines in file {extracted_items_file}, written empty load job file"
                )

        return schema_updates



from .file_storage import FileStorage
from .versioned_storage import VersionedStorage
from .schema_storage import SchemaStorage
from .live_schema_storage import LiveSchemaStorage
from .normalize_storage import NormalizeStorage
from .load_package import (
    ParsedLoadJobFileName,
    LoadJobInfo,
    LoadPackageInfo,
    PackageStorage,
    TPackageJobState,
    create_load_id,
)
from .data_item_storage import DataItemStorage
from .load_storage import LoadStorage
from .configuration import (
    LoadStorageConfiguration,
    NormalizeStorageConfiguration,
    SchemaStorageConfiguration,
    TSchemaFileFormat,
    FilesystemConfiguration,
    WithLocalFiles,
    FilesystemConfigurationWithLocalFiles,
)
from .fsspec_filesystem import fsspec_from_config, fsspec_filesystem


__all__ = [
    "FileStorage",
    "VersionedStorage",
    "SchemaStorage",
    "LiveSchemaStorage",
    "NormalizeStorage",
    "LoadStorage",
    "DataItemStorage",
    "LoadStorageConfiguration",
    "NormalizeStorageConfiguration",
    "SchemaStorageConfiguration",
    "TSchemaFileFormat",
    "FilesystemConfiguration",
    "WithLocalFiles",
    "FilesystemConfigurationWithLocalFiles",
    "ParsedLoadJobFileName",
    "LoadJobInfo",
    "LoadPackageInfo",
    "PackageStorage",
    "TPackageJobState",
    "create_load_id",
    "fsspec_from_config",
    "fsspec_filesystem",
]



from typing import Any, Optional, Sequence, Tuple, Type, Union, Dict, TYPE_CHECKING

from dlt.common import logger
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.destination.configuration import CsvFormatConfiguration
from dlt.common.data_writers.escape import escape_postgres_identifier, escape_postgres_literal
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema.typing import TColumnSchema, TColumnType, TTableSchema
from dlt.common.typing import TLoaderFileFormat
from dlt.common.wei import EVM_DECIMAL_PRECISION
from dlt.destinations._adbc_jobs import make_adbc_parquet_file_format_selector
from dlt.destinations.impl.postgres.configuration import (
    PostgresCredentials,
    PostgresClientConfiguration,
)
from dlt.destinations.impl.postgres.postgres_adapter import GEOMETRY_HINT, SRID_HINT
from dlt.destinations.type_mapping import TypeMapperImpl

if TYPE_CHECKING:
    from dlt.destinations.impl.postgres.postgres import PostgresClient


class PostgresTypeMapper(TypeMapperImpl):
    sct_to_unbound_dbt = {
        "json": "jsonb",
        "text": "varchar",
        "double": "double precision",
        "bool": "boolean",
        "date": "date",
        "bigint": "bigint",
        "binary": "bytea",
        "timestamp": "timestamp with time zone",
        "time": "time without time zone",
    }

    sct_to_dbt = {
        "text": "varchar(%i)",
        "timestamp": "timestamp (%i) with time zone",
        "decimal": "numeric(%i,%i)",
        "time": "time (%i) without time zone",
        "wei": "numeric(%i,%i)",
    }

    dbt_to_sct = {
        "varchar": "text",
        "jsonb": "json",
        "double precision": "double",
        "boolean": "bool",
        "timestamp with time zone": "timestamp",
        "timestamp without time zone": "timestamp",
        "date": "date",
        "bigint": "bigint",
        "bytea": "binary",
        "numeric": "decimal",
        "time without time zone": "time",
        "character varying": "text",
        "smallint": "bigint",
        "integer": "bigint",
        "geometry": "text",
    }

    def to_db_integer_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        precision = column.get("precision")
        if precision is None:
            return "bigint"
        # Precision is number of bits
        if precision <= 16:
            return "smallint"
        elif precision <= 32:
            return "integer"
        elif precision <= 64:
            return "bigint"
        raise TerminalValueError(
            f"bigint with `{precision=:}` can't be mapped to PostgreSQL integer type"
        )

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        column_name = column.get("name")
        table_name = table.get("name")
        timezone = column.get("timezone")
        precision = column.get("precision")

        if timezone is None and precision is None:
            return None

        timestamp = "timestamp"

        # append precision if specified and valid
        if precision is not None:
            if 0 <= precision <= self.capabilities.timestamp_precision:
                timestamp += f" ({precision})"
            else:
                raise TerminalValueError(
                    f"Postgres doesn't support {precision=:} for timestamp column `{column_name}`"
                    f" in table `{table_name}`"
                )

        # append timezone part
        if timezone is None or timezone:  # timezone True and None
            timestamp += " with time zone"
        else:  # timezone is explicitly False
            timestamp += " without time zone"

        return timestamp

    def from_destination_type(
        self, db_type: str, precision: Optional[int] = None, scale: Optional[int] = None
    ) -> TColumnType:
        if db_type == "numeric" and (precision, scale) == self.capabilities.wei_precision:
            return dict(data_type="wei")
        if db_type.startswith("geometry"):
            return dict(data_type="text")
        if db_type.startswith("json"):
            return dict(data_type="json")
        return super().from_destination_type(db_type, precision, scale)

    def to_destination_type(self, column: TColumnSchema, table: PreparedTableSchema) -> str:
        if column.get(GEOMETRY_HINT):
            srid = column.get(SRID_HINT, 4326)
            return f"geometry(Geometry, {srid})"
        if column["data_type"] == "json" and table.get("file_format") == "parquet":
            return "json"  # adbc is not able to copy into JSONB columns
        return super().to_destination_type(column, table)


def get_adbc_driver_location() -> str:
    """Detects driver location if PyPI driver package is installed, otherwise falls back to dbc
    driver name.
    """
    # TODO: move to the (future) ADBC helper module
    try:
        from adbc_driver_postgresql import _driver_path

        # use driver from installed dependency
        return _driver_path()  # type: ignore[no-any-return]
    except Exception:
        return "postgresql"


class postgres(Destination[PostgresClientConfiguration, "PostgresClient"]):
    spec = PostgresClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        # https://www.postgresql.org/docs/current/limits.html
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "insert_values"
        caps.supported_loader_file_formats = ["insert_values", "csv", "parquet", "model"]
        caps.loader_file_format_selector = make_adbc_parquet_file_format_selector(
            get_adbc_driver_location(),
            "https://dlthub.com/docs/dlt-ecosystem/destinations/postgres#fast-loading-with-arrow-tables-and-parquet",
            prefer_parquet=False,
        )
        caps.preferred_staging_file_format = None
        caps.supported_staging_file_formats = []
        caps.type_mapper = PostgresTypeMapper
        caps.escape_identifier = escape_postgres_identifier
        # postgres has case sensitive identifiers but by default
        # it folds them to lower case which makes them case insensitive
        # https://stackoverflow.com/questions/20878932/are-postgresql-column-names-case-sensitive
        caps.casefold_identifier = str.lower
        caps.has_case_sensitive_identifiers = True
        caps.escape_literal = escape_postgres_literal
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (2 * EVM_DECIMAL_PRECISION, EVM_DECIMAL_PRECISION)
        caps.max_identifier_length = 63
        caps.max_column_identifier_length = 63
        caps.max_query_length = 32 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 1024 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = True
        caps.supported_merge_strategies = ["delete-insert", "upsert", "scd2"]
        caps.supported_replace_strategies = [
            "truncate-and-insert",
            "insert-from-staging",
            "staging-optimized",
        ]
        caps.sqlglot_dialect = "postgres"

        return caps

    @property
    def client_class(self) -> Type["PostgresClient"]:
        from dlt.destinations.impl.postgres.postgres import PostgresClient

        return PostgresClient

    def __init__(
        self,
        credentials: Union[PostgresCredentials, Dict[str, Any], str] = None,
        create_indexes: bool = True,
        csv_format: Optional[CsvFormatConfiguration] = None,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Postgres destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[PostgresCredentials, Dict[str, Any], str], optional): Credentials to connect to the postgres database. Can be an instance of `PostgresCredentials` or
                a connection string in the format `postgres://user:password@host:port/database`
            create_indexes (bool, optional): Should unique indexes be created
            csv_format (Optional[CsvFormatConfiguration]): Formatting options for csv file format
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            create_indexes=create_indexes,
            csv_format=csv_format,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


postgres.register()



import posixpath
import os
import time as _time
import orjson
import base64
from contextlib import contextmanager
from types import TracebackType
from typing import (
    ClassVar,
    List,
    Type,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    Sequence,
    cast,
    Any,
    Dict,
    TYPE_CHECKING,
)
from fsspec import AbstractFileSystem
from tenacity import (
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential,
    RetryCallState,
)

from dlt.common import logger, time, json, pendulum
from dlt.common.destination.utils import resolve_merge_strategy, resolve_replace_strategy
from dlt.common.metrics import LoadJobMetrics
from dlt.common.schema.exceptions import TableNotFound
from dlt.common.schema.typing import (
    C_DLT_LOAD_ID,
    C_DLT_LOADS_TABLE_LOAD_ID,
    TTableFormat,
    TTableSchemaColumns,
)
from dlt.common.storages.exceptions import (
    CurrentLoadPackageStateNotAvailable,
    NoMigrationPathException,
    UnsupportedStorageVersionException,
)
from dlt.common.storages.fsspec_filesystem import glob_files
from dlt.common.time import ensure_pendulum_datetime_utc
from dlt.common.typing import ConfigValue, DictStrAny
from dlt.common.schema import Schema, TSchemaTables
from dlt.common.schema.utils import get_columns_names_with_prop, is_nested_table
from dlt.common.storages import FileStorage, fsspec_from_config
from dlt.common.storages.load_package import (
    LoadJobInfo,
    ParsedLoadJobFileName,
    TPipelineStateDoc,
    load_package_state as current_load_package,
)
from dlt.destinations.sql_client import WithSqlClient, SqlClientBase
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    FollowupJobRequest,
    PreparedTableSchema,
    SupportsOpenTables,
    TLoadJobState,
    RunnableLoadJob,
    JobClientBase,
    HasFollowupJobs,
    WithStagingDataset,
    WithStateSync,
    StorageSchemaInfo,
    StateInfo,
    LoadJob,
)
from dlt.common.destination.exceptions import (
    TableFormatNotSupported,
    WriteDispositionNotSupported,
    DestinationUndefinedEntity,
    OpenTableCatalogNotSupported,
    OpenTableFormatNotSupported,
)
from dlt.common.schema.exceptions import SchemaCorruptedException
from dlt.destinations.job_impl import (
    ReferenceFollowupJobRequest,
    FinalizedLoadJob,
    FinalizedLoadJobWithFollowupJobs,
)
from dlt.destinations.impl.filesystem.configuration import (
    FilesystemDestinationClientConfiguration,
    HfFilesystemDestinationClientConfiguration,
)
from dlt.destinations import path_utils
from dlt.destinations.fs_client import FSClientBase
from dlt.destinations.utils import (
    verify_schema_merge_disposition,
    verify_schema_replace_disposition,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.filesystem.iceberg_adapter import PartitionSpec

CURRENT_VERSION: int = 2
SUPPORTED_VERSIONS: set[int] = {1, CURRENT_VERSION}

INIT_FILE_NAME = "init"
FILENAME_SEPARATOR = "__"


class FilesystemLoadJob(RunnableLoadJob):
    def __init__(
        self,
        file_path: str,
    ) -> None:
        super().__init__(file_path)
        self._job_client: FilesystemClient = None

    def run(self) -> None:
        self.__is_local_filesystem = self._job_client.config.is_local_filesystem
        # We would like to avoid failing for local filesystem where
        # deeply nested directory will not exist before writing a file.
        # It `auto_mkdir` is disabled by default in fsspec so we made some
        # trade offs between different options and decided on this.
        remote_path = self.make_remote_path()
        if (
            not self._job_client.config.as_staging_destination
            and self._job_client.storage_versions[0] == 1
            and remote_path.endswith(".gz")
        ):
            remote_path = remote_path[:-3]
        if self.__is_local_filesystem:
            # use os.path for local file name
            self._job_client.fs_client.makedirs(os.path.dirname(remote_path), exist_ok=True)
        self._job_client.fs_client.put_file(self._file_path, remote_path)

    @property
    def load_package_timestamp(self) -> Optional[pendulum.DateTime]:
        # package state is optional
        try:
            return ensure_pendulum_datetime_utc(current_load_package()["state"]["created_at"])
        except CurrentLoadPackageStateNotAvailable:
            return None

    def make_remote_file_path(self, file_name: str) -> str:
        """Returns path on remote filesystem to move file to, without scheme and dataset."""
        return path_utils.create_path(
            layout=self._job_client.config.layout,
            file_name=file_name,
            schema_name=self._job_client.schema.name,
            load_id=self._load_id,
            current_datetime=self._job_client.config.current_datetime,
            load_package_timestamp=self.load_package_timestamp,
            extra_placeholders=self._job_client.config.extra_placeholders,
        )

    def make_remote_path(self) -> str:
        """Returns path on remote filesystem to move file(s) to, without scheme, but with dataset.

        For local filesystem a native path is used.
        """
        remote_file_path = self.make_remote_file_path(self._file_name)
        # pick local filesystem pathlib or posix for buckets
        pathlib = self._job_client.config.pathlib
        # path.join does not normalize separators and available
        # normalization functions are very invasive and may strip the trailing separator
        return pathlib.join(  # type: ignore[no-any-return]
            self._job_client.dataset_path,
            path_utils.normalize_path_sep(pathlib, remote_file_path),
        )

    def make_remote_url(self) -> str:
        """Returns path on a remote filesystem as a full url including scheme."""
        return self._job_client.make_remote_url(self.make_remote_path())

    def metrics(self) -> Optional[LoadJobMetrics]:
        m = super().metrics()
        # job client not available before run_managed called
        if not self._job_client:
            return m
        return m._replace(remote_url=self.make_remote_url())


class ReferenceFollowupJob(FilesystemLoadJob):
    def __init__(self, file_path: str) -> None:
        super().__init__(file_path=file_path)
        self.file_paths = ReferenceFollowupJobRequest.resolve_references(self._file_path)

    def make_remote_path(self) -> str:
        return self._job_client.get_table_dir(self.load_table_name)


class TableFormatLoadFilesystemJob(ReferenceFollowupJob):
    @property
    def arrow_dataset(self) -> Any:
        from dlt.common.libs.pyarrow import pyarrow

        return pyarrow.dataset.dataset(self.file_paths)

    @property
    def _partition_columns(self) -> List[str]:
        return get_columns_names_with_prop(self._load_table, "partition")


class DeltaLoadFilesystemJob(TableFormatLoadFilesystemJob):
    def run(self) -> None:
        # create Arrow dataset from Parquet files
        from dlt.common.libs.pyarrow import pyarrow as pa
        from dlt.common.libs.deltalake import (
            write_delta_table,
            merge_delta_table,
            DeltaTable,
            deltalake_storage_options,
        )

        logger.info(
            f"Will copy file(s) {self.file_paths} to delta table {self.make_remote_url()} [arrow"
            f" buffer: {pa.total_allocated_bytes()}]"
        )
        source_ds = self.arrow_dataset
        storage_options = deltalake_storage_options(
            self._job_client.config.credentials,
            self._job_client.config.deltalake_storage_options,
        )
        try:
            delta_table: DeltaTable = self._job_client.load_open_table(
                "delta", self.load_table_name
            )
        except DestinationUndefinedEntity:
            delta_table = None

        with source_ds.scanner().to_reader() as arrow_rbr:  # RecordBatchReader
            if self._load_table["write_disposition"] == "merge" and delta_table is not None:
                merge_delta_table(
                    table=delta_table,
                    data=arrow_rbr,
                    schema=self._load_table,
                    load_table_name=self.load_table_name,
                    streamed_exec=self._job_client.config.deltalake_streamed_exec,
                )
            else:
                location = self._job_client.get_open_table_location("delta", self.load_table_name)
                write_delta_table(
                    table_or_uri=location if delta_table is None else delta_table,
                    data=arrow_rbr,
                    write_disposition=self._load_table["write_disposition"],
                    partition_by=self._partition_columns,
                    storage_options=storage_options,
                    configuration=self._job_client.config.deltalake_configuration,
                )
        # release memory ASAP by deleting objects explicitly
        del source_ds
        del delta_table
        logger.info(
            f"Copied {self.file_paths} to delta table {self.make_remote_url()} [arrow buffer:"
            f" {pa.total_allocated_bytes()}]"
        )


class IcebergLoadFilesystemJob(TableFormatLoadFilesystemJob):
    def run(self) -> None:
        from dlt.common.libs.pyiceberg import (
            write_iceberg_table,
            merge_iceberg_table,
            create_table,
        )
        from dlt.destinations.impl.filesystem.iceberg_partition_spec import (
            build_iceberg_partition_spec,
        )

        try:
            table = self._job_client.load_open_table(
                "iceberg",
                self.load_table_name,
                schema=self.arrow_dataset.schema,
            )
        except DestinationUndefinedEntity:
            from dlt.destinations.impl.filesystem.iceberg_adapter import TABLE_PROPERTIES_HINT

            location = self._job_client.get_open_table_location("iceberg", self.load_table_name)
            table_id = f"{self._job_client.dataset_name}.{self.load_table_name}"

            spec_list = self._get_partition_spec_list()
            # merge config-level defaults with per-table adapter overrides
            config_properties = self._job_client.config.iceberg_table_properties
            adapter_properties: Optional[Dict[str, str]] = self._load_table.get(
                TABLE_PROPERTIES_HINT
            )  # type: ignore[assignment]
            if config_properties or adapter_properties:
                properties = {**(config_properties or {}), **(adapter_properties or {})}
            else:
                properties = None

            if spec_list:
                partition_spec, iceberg_schema = build_iceberg_partition_spec(
                    self.arrow_dataset.schema, spec_list
                )
                create_table(
                    self._job_client.get_open_table_catalog("iceberg"),
                    table_id,
                    table_location=location,
                    schema=iceberg_schema,
                    partition_spec=partition_spec,
                    properties=properties,
                )
            else:
                create_table(
                    self._job_client.get_open_table_catalog("iceberg"),
                    table_id,
                    table_location=location,
                    schema=self.arrow_dataset.schema,
                    properties=properties,
                )
            # run again with created table
            self.run()
            return

        if self._load_table["write_disposition"] == "merge" and table is not None:
            merge_iceberg_table(
                table=table,
                data=self.arrow_dataset.to_table(),
                schema=self._load_table,
                load_table_name=self.load_table_name,
            )
        else:
            write_iceberg_table(
                table=table,
                data=self.arrow_dataset.to_table(),
                write_disposition=self._load_table["write_disposition"],
            )

    def _get_partition_spec_list(self) -> List["PartitionSpec"]:
        """Resolve partition specs. Combines legacy partition columns (identity transform)
        with partition hints. Validates that identity partitions are not duplicated.
        """
        from dlt.destinations.impl.filesystem.iceberg_adapter import (
            parse_partition_hints,
            create_identity_specs,
        )

        legacy_columns = self._partition_columns

        hint_specs = parse_partition_hints(self._load_table)

        for spec in hint_specs:
            if spec.transform == "identity" and spec.source_column in legacy_columns:
                raise SchemaCorruptedException(
                    self._schema.name,
                    f"Column '{spec.source_column}' is defined both as a partition column "
                    "and in partition hints.",
                )

        identity_specs = create_identity_specs(legacy_columns)
        return identity_specs + hint_specs


class HfFilesystemUploadJob(HasFollowupJobs, FilesystemLoadJob):
    """Pre-uploads a single file to HF LFS storage without creating a commit.

    File content is uploaded to HF's content-addressed storage via preupload_lfs_files.
    The actual git commit is created later by HfFilesystemCommitJob once all uploads
    in the table chain complete.
    """

    _job_client: "HfFilesystemClient"

    def run(self) -> None:
        from huggingface_hub import CommitOperationAdd

        operation = CommitOperationAdd(
            path_in_repo=self.make_remote_file_path(self._file_name),
            path_or_fileobj=self._file_path,
        )
        start = _time.monotonic()
        self._job_client.hf_api.preupload_lfs_files(
            repo_id=self._job_client.repo_id,
            additions=[operation],
            repo_type="dataset",
        )
        elapsed = _time.monotonic() - start
        logger.info(
            f"HF pre-upload of {self._file_name} to {self._job_client.repo_id}"
            f" completed in {elapsed:.1f}s"
        )


def _is_hf_conflict(exc: BaseException) -> bool:
    """Return True for HF Hub commit conflict errors (409 or 412)."""
    from huggingface_hub.errors import HfHubHTTPError

    return (
        isinstance(exc, HfHubHTTPError)
        and hasattr(exc, "response")
        and exc.response is not None
        and exc.response.status_code in (409, 412)
    )


def _log_hf_retry(retry_state: RetryCallState) -> None:
    """Log warning on each tenacity retry for HF commit conflicts."""
    exc = retry_state.outcome.exception()
    wait = retry_state.next_action.sleep if retry_state.next_action else 0
    logger.warning(
        f"HF commit conflict on attempt {retry_state.attempt_number},"
        f" retrying after {wait:.1f}s: {exc}"
    )


class HfFilesystemCommitJob(ReferenceFollowupJob):
    """Commits pre-uploaded files to a HF dataset repo.

    Files are already in HF's content-addressed storage (uploaded by HfFilesystemUploadJob).
    create_commit internally detects pre-uploaded blobs and only creates the git commit.
    Retries on 409/412 commit conflicts with exponential backoff via tenacity.
    Other errors propagate to dlt load engine for dlt-level retry.
    """

    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: HfFilesystemClient = None

    @retry(
        retry=retry_if_exception(_is_hf_conflict),
        wait=wait_exponential(multiplier=1, max=30),
        stop=stop_after_attempt(5),
        reraise=True,
        before_sleep=_log_hf_retry,
    )
    def run(self) -> None:
        from huggingface_hub import CommitOperationAdd

        operations: List[Any] = [
            CommitOperationAdd(path_in_repo=self.make_path_in_repo(path), path_or_fileobj=path)
            for path in self.file_paths
        ]

        # for replace disposition, delete existing table files in the same atomic commit
        delete_ops = self._get_delete_operations()
        if delete_ops:
            operations = delete_ops + operations

        table_names = list(
            dict.fromkeys(
                ParsedLoadJobFileName.parse(self._job_client.pathlib.basename(path)).table_name
                for path in self.file_paths
            )
        )
        tables_str = ", ".join(table_names)

        start = _time.monotonic()
        self._job_client.hf_api.create_commit(
            repo_id=self._job_client.repo_id,
            operations=operations,
            commit_message=f"Load {self._load_id}: add files to {tables_str}",
            repo_type="dataset",
        )
        elapsed = _time.monotonic() - start
        n_adds = len(self.file_paths)
        n_deletes = len(delete_ops) if delete_ops else 0
        logger.info(
            f"HF commit to {self._job_client.repo_id} for {tables_str}"
            f" ({n_adds} added, {n_deletes} deleted) completed in {elapsed:.1f}s"
        )
        self._job_client.fs_client.invalidate_cache()

    def _get_delete_operations(self) -> List[Any]:
        """Return CommitOperationDelete ops for tables with replace write disposition."""
        table_names_to_replace = []
        for path in self.file_paths:
            file_name = self._job_client.pathlib.basename(path)
            parsed = ParsedLoadJobFileName.parse(file_name)
            table = self._job_client.prepare_load_table(parsed.table_name)
            if (
                table["write_disposition"] == "replace"
                and parsed.table_name not in table_names_to_replace
            ):
                table_names_to_replace.append(parsed.table_name)

        return self._job_client._collect_delete_operations(table_names_to_replace)

    def make_path_in_repo(self, file_path: str) -> str:
        """Returns path relative to repo root, without namespace and repo name."""
        file_name = self._job_client.pathlib.basename(file_path)
        return self.make_remote_file_path(file_name)


class FilesystemLoadJobWithFollowup(HasFollowupJobs, FilesystemLoadJob):
    def create_followup_jobs(self, final_state: TLoadJobState) -> List[FollowupJobRequest]:
        jobs = super().create_followup_jobs(final_state)
        if final_state == "completed":
            ref_job = ReferenceFollowupJobRequest(
                original_file_name=self.file_name(),
                remote_paths=[self._job_client.make_remote_url(self.make_remote_path())],
            )
            jobs.append(ref_job)
        return jobs


class FilesystemClient(
    FSClientBase,
    WithSqlClient,
    JobClientBase,
    WithStagingDataset,
    WithStateSync,
    SupportsOpenTables,
):
    fs_client: AbstractFileSystem
    # a path (without the scheme) to a location in the bucket where dataset is present
    bucket_path: str
    # name of the dataset
    dataset_name: str
    batch_table_chain: ClassVar[bool] = False
    """If True, uses single reference followup job for entire table chain. If False, uses separate reference followup job for each table in chain."""

    def __init__(
        self,
        schema: Schema,
        config: FilesystemDestinationClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(schema, config, capabilities)
        self.fs_client, fs_path = fsspec_from_config(config)
        self.is_local_filesystem = config.is_local_filesystem
        self.bucket_path = (
            config.make_local_path(config.bucket_url) if self.is_local_filesystem else fs_path
        )
        # pick local filesystem pathlib or posix for buckets
        self.pathlib = os.path if self.is_local_filesystem else posixpath

        self.config: FilesystemDestinationClientConfiguration = config
        # verify files layout. we need {table_name} and only allow {schema_name} before it, otherwise tables
        # cannot be replaced and we cannot initialize folders consistently
        self.table_prefix_layout = path_utils.get_table_prefix_layout(config.layout)
        self.dataset_name = self.config.normalize_dataset_name(self.schema)
        self._sql_client: SqlClientBase[Any] = None
        # iceberg catalog
        self._catalog: Any = None

        # storage versions
        self._cached_initial_storage_version: int = None
        self._cached_storage_version: int = None

    @property
    def sql_client_class(self) -> Type[SqlClientBase[Any]]:
        from dlt.destinations.impl.filesystem.sql_client import FilesystemSqlClient

        return FilesystemSqlClient

    @property
    def sql_client(self) -> SqlClientBase[Any]:
        # we use an inner import here, since the sql client depends on duckdb and will
        # only be used for read access on data, some users will not need the dependency
        from dlt.destinations.impl.filesystem.sql_client import FilesystemSqlClient

        if not self._sql_client:
            self._sql_client = FilesystemSqlClient(self, self.dataset_name)
        return self._sql_client

    @sql_client.setter
    def sql_client(self, client: SqlClientBase[Any]) -> None:
        self._sql_client = client

    @property
    def storage_versions(self) -> Tuple[int, int]:
        """Returns cached storage versions, loading it once from filesystem if not already cached"""
        if self._cached_initial_storage_version is None or self._cached_storage_version is None:
            self._cached_initial_storage_version, self._cached_storage_version = (
                self.get_storage_versions()
            )
        return self._cached_initial_storage_version, self._cached_storage_version

    @property
    def init_file_path(self) -> str:
        """Returns the path to the init file for the current dataset"""
        return str(self.pathlib.join(self.dataset_path, INIT_FILE_NAME))

    def create_dataset(self) -> None:
        """Creates empty dataset directory."""
        self.fs_client.makedirs(self.dataset_path)

    def drop_dataset(self) -> None:
        """Removes dataset directory with all its content."""
        self.fs_client.rm(self.dataset_path, recursive=True)

    def drop_storage(self) -> None:
        if self.is_storage_initialized():
            self.drop_dataset()

    @property
    def dataset_path(self) -> str:
        """A path within a bucket to tables in a dataset
        NOTE: dataset_name changes if with_staging_dataset is active
        """
        return self.pathlib.join(self.bucket_path, self.dataset_name, "")  # type: ignore[no-any-return]

    @contextmanager
    def with_staging_dataset(self) -> Iterator["FilesystemClient"]:
        current_dataset_name = self.dataset_name
        try:
            _, self.dataset_name = WithStagingDataset.create_dataset_names(self.schema, self.config)
            yield self
        finally:
            # restore previous dataset name
            self.dataset_name = current_dataset_name

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        # clean up existing files for tables selected for truncating
        if truncate_tables and self.fs_client.isdir(self.dataset_path):
            # get all dirs with table data to delete. the table data are guaranteed to be files in those folders
            # TODO: when we do partitioning it is no longer the case and we may remove folders below instead
            logger.info(f"Will truncate tables {truncate_tables}")
            self.truncate_tables(list(truncate_tables))

        # check if init file already exists
        if self.fs_client.exists(self.init_file_path):
            current_version = self.get_storage_versions()[0]

            # check if migration is needed
            if current_version != CURRENT_VERSION:
                if current_version > CURRENT_VERSION:
                    # version cannot be downgraded
                    raise NoMigrationPathException(
                        self.dataset_path, current_version, current_version, CURRENT_VERSION
                    )

                # migrate and verify
                self.migrate_storage(current_version, CURRENT_VERSION)
                migrated_version = self.get_storage_versions()[1]
                if migrated_version != CURRENT_VERSION:
                    raise ValueError(
                        f"Migration failed: expected {CURRENT_VERSION}, got {migrated_version}"
                    )

            return

        self.create_dataset()

        # we set the version to "2" to indicate that .gz extension is added by default
        # version "1" corresponds to an empty init file and indicates that .gz is not added to compressed files
        self.fs_client.write_text(
            self.init_file_path,
            json.dumps({"initial_version": CURRENT_VERSION, "current_version": CURRENT_VERSION}),
            encoding="utf-8",
        )

    def migrate_storage(self, from_version: int, to_version: int) -> None:
        """Migrate storage from one version to another"""

        if from_version == 1 and from_version < to_version:
            # Migration from version 1 to 2,
            # which is basically replacing the empty init file with:
            # {"initial_version": 1, "current_version": 2}
            self.fs_client.write_text(
                self.init_file_path,
                json.dumps({"initial_version": 1, "current_version": 2}),
                encoding="utf-8",
            )

        # Reset cached current version, initial version may stay
        self._cached_storage_version = None

    def get_storage_versions(self) -> Tuple[int, int]:
        """Returns initial and current storage versions.
        1. If the init file is empty, we assume legacy version 1 where .gz extension was not added to compressed files.
        2. For any other non-empty content we parse it as json, expect version key to have a supported value.
        """
        version_info_str = self.fs_client.read_text(self.init_file_path, encoding="utf-8")

        # If empty, both initial and current versions are 1
        if not version_info_str.strip():
            return 1, 1

        try:
            version_dict: Dict[str, int] = json.loads(version_info_str)
            initial_version: int = version_dict["initial_version"]
            current_version: int = version_dict["current_version"]
        except (KeyError, orjson.JSONDecodeError) as exc:
            raise ValueError(
                f"Invalid content in {self.init_file_path}: {version_info_str!r}"
            ) from exc

        if initial_version not in SUPPORTED_VERSIONS or current_version not in SUPPORTED_VERSIONS:
            raise UnsupportedStorageVersionException(
                self.dataset_path,
                initial_version,
                current_version,
                SUPPORTED_VERSIONS,
            )

        return initial_version, current_version

    def drop_tables(self, *tables: str, delete_schema: bool = True) -> None:
        self.truncate_tables(list(tables))
        if not delete_schema:
            return
        # Delete all stored schemas
        for filename, fileparts in self._iter_stored_schema_files():
            if fileparts[0] == self.schema.name:
                self._delete_file(filename)

    def get_storage_tables(
        self, table_names: Iterable[str]
    ) -> Iterable[Tuple[str, TTableSchemaColumns]]:
        """Yields tables that have files in storage, returns columns from current schema"""
        for table_name in table_names:
            table_dir = self.get_table_dir(table_name)
            if (
                # TODO: isdir is sufficient if table_dir == table prefix
                #   since this method is used currently only for tests we do not need to improve it
                self.fs_client.isdir(table_dir)
                and len(self.list_table_files(table_name)) > 0
            ):
                if table_name in self.schema.tables:
                    yield (table_name, self.schema.get_table_columns(table_name))
                else:
                    yield (table_name, {"_column": {}})
            else:
                # if no columns we assume that table does not exist
                yield (table_name, {})

    def get_storage_table(self, table_name: str) -> Tuple[str, TTableSchemaColumns]:
        return list(self.get_storage_tables([table_name]))[0]

    def truncate_tables(self, table_names: List[str]) -> None:
        """Truncate a set of regular tables with given `table_names`"""
        table_dirs = set(self.get_table_dirs(table_names))
        table_prefixes = [self.get_table_prefix(t) for t in table_names]
        for table_dir in table_dirs:
            if self.fs_client.exists(table_dir):
                for table_file in self.list_files_with_prefixes(table_dir, table_prefixes):
                    # NOTE: deleting in chunks on s3 does not raise on access denied, file non existing and probably other errors
                    # print(f"DEL {table_file}")
                    try:
                        self._delete_file(table_file)
                    except FileNotFoundError:
                        logger.info(
                            f"Directory or path to truncate tables {table_names} does not exist but"
                            " it should have been created previously!"
                        )

    def _delete_file(self, file_path: str) -> None:
        try:
            # NOTE: must use rm_file to get errors on delete
            self.fs_client.rm_file(file_path)
        except FileNotFoundError:
            # File doesn't exist - deletion goal already achieved
            return
        except Exception as e:
            # OneLake API quirk: returns HTTP 200 instead of HTTP 202 for delete operations
            # Azure SDK treats this as an error, but the delete actually succeeded
            if "Operation returned an invalid status 'OK'" in str(e):
                # Verify deletion succeeded by checking if file still exists
                if not self.fs_client.exists(file_path):
                    return  # Delete succeeded despite the error
            # If it's NotImplementedError, try the fallback rm() method
            if isinstance(e, NotImplementedError):
                try:
                    self.fs_client.rm(file_path)
                except FileNotFoundError:
                    # File doesn't exist - deletion goal already achieved
                    return
                except Exception as rm_e:
                    # OneLake API quirk: returns HTTP 200 instead of HTTP 202 for delete operations
                    # Azure SDK treats this as an error, but the delete actually succeeded
                    if "Operation returned an invalid status 'OK'" in str(rm_e):
                        # Verify deletion succeeded by checking if file still exists
                        if not self.fs_client.exists(file_path):
                            return  # Delete succeeded despite the error
                    raise
                if self.fs_client.exists(file_path):
                    raise FileExistsError(file_path)
            else:
                raise

    def verify_schema(
        self, only_tables: Iterable[str] = None, new_jobs: Iterable[ParsedLoadJobFileName] = None
    ) -> List[PreparedTableSchema]:
        loaded_tables = super().verify_schema(only_tables, new_jobs)
        # TODO: finetune verify_schema_merge_disposition ie. hard deletes are not supported
        if exceptions := verify_schema_merge_disposition(
            self.schema, loaded_tables, self.capabilities, warnings=True
        ):
            # filesystem falls back to append when merge is not supported
            filtered = []
            for exception in exceptions:
                if isinstance(exception, WriteDispositionNotSupported):
                    logger.warning(str(exception))
                else:
                    filtered.append(exception)
                    logger.error(str(exception))
            if filtered:
                raise filtered[0]
        if exceptions := verify_schema_replace_disposition(
            self.schema,
            loaded_tables,
            self.capabilities,
            self.config.replace_strategy,
            warnings=True,
        ):
            for exception in exceptions:
                logger.error(str(exception))
            raise exceptions[0]
        if exceptions := self.verify_schema_table_format(loaded_tables):
            for exception in exceptions:
                logger.error(str(exception))
            raise exceptions[0]
        return loaded_tables

    def verify_schema_table_format(
        self, load_tables: Sequence[PreparedTableSchema]
    ) -> List[Exception]:
        exception_log: List[Exception] = []
        for table in load_tables:
            # from now on validate only top level tables
            if is_nested_table(table):
                continue
            table_format = table.get("table_format")
            if table_format and table_format not in self.capabilities.supported_table_formats:
                message = f"supported table formats are {self.capabilities.supported_table_formats}"
                if self.config.protocol == "hf":
                    message = "the `hf` protocol does not support table formats"
                exception_log.append(TableFormatNotSupported(table_format, table["name"], message))
        return exception_log

    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> TSchemaTables:
        applied_update = super().update_stored_schema(only_tables, expected_update)

        # don't store schema when used as staging
        if not self.config.as_staging_destination:
            # check if schema with hash exists
            current_hash = self.schema.stored_version_hash
            if not self._get_stored_schema_by_hash_or_newest(current_hash):
                logger.info(
                    f"Schema with hash {self.schema.stored_version_hash} not found in the storage."
                    " upgrading"
                )
                # create destination dirs for all tables
                # TODO: find only tables with changes
                table_names = only_tables or self.schema.tables.keys()
                dirs_to_create = self.get_table_dirs(table_names)
                for _, directory in zip(table_names, dirs_to_create):
                    self.fs_client.makedirs(directory, exist_ok=True)
                self._update_schema_in_storage(self.schema)

        # we assume that expected_update == applied_update so table schemas in dest were not
        # externally changed
        return applied_update

    def prepare_load_table(self, table_name: str) -> PreparedTableSchema:
        table = super().prepare_load_table(table_name)
        if self.config.as_staging_destination:
            if table["write_disposition"] in ("merge", "replace"):
                table["write_disposition"] = "append"
            table.pop("table_format", None)
        merge_strategy = resolve_merge_strategy(self.schema.tables, table, self.capabilities)
        if table["write_disposition"] == "merge":
            if merge_strategy is None:
                # no supported merge strategies, fall back to append
                logger.warning(f"Falling back to `append` on `{table_name}`.")
                table["write_disposition"] = "append"
            else:
                table["x-merge-strategy"] = merge_strategy  # type: ignore[typeddict-unknown-key]
        if table["write_disposition"] == "replace":
            replace_strategy = resolve_replace_strategy(
                table, self.config.replace_strategy, self.capabilities
            )
            assert replace_strategy, f"Must be able to get replace strategy for {table_name}"
            table["x-replace-strategy"] = replace_strategy  # type: ignore[typeddict-unknown-key]
        if self.is_dlt_table(table["name"]):
            table.pop("table_format", None)
        return table

    def get_table_dir(self, table_name: str, remote: bool = False) -> str:
        """Returns a directory containing table files, ending with separator.
        Note that many tables can share the same table dir
        """
        # dlt tables do not respect layout (for now)
        table_prefix = self.get_table_prefix(table_name)
        table_dir: str = self.pathlib.dirname(table_prefix) + self.pathlib.sep
        if remote:
            table_dir = self.make_remote_url(table_dir)
        return table_dir

    def get_table_prefix(self, table_name: str) -> str:
        """For table prefixes that are folders, trailing separator will be preserved"""
        # dlt tables do not respect layout (for now)
        if self.is_dlt_table(table_name):
            # dlt tables get layout where each tables is a folder
            # it is crucial to append and keep "/" at the end
            table_prefix = self.pathlib.join(table_name, "")
        else:
            table_prefix = self.table_prefix_layout.format(
                schema_name=self.schema.name, table_name=table_name
            )
        return self.pathlib.join(  # type: ignore[no-any-return]
            self.dataset_path, path_utils.normalize_path_sep(self.pathlib, table_prefix)
        )

    def is_dlt_table(self, table_name: str) -> bool:
        return table_name.startswith(self.schema._dlt_tables_prefix)

    def get_table_dirs(self, table_names: Iterable[str], remote: bool = False) -> List[str]:
        """Gets directories where table data is stored."""
        return [self.get_table_dir(t, remote=remote) for t in table_names]

    def list_table_files(self, table_name: str) -> List[str]:
        """gets list of files associated with one table"""
        table_dir = self.get_table_dir(table_name)
        # we need the table prefix so we separate table files if in the same folder
        table_prefix = self.get_table_prefix(table_name)
        try:
            return self.list_files_with_prefixes(table_dir, [table_prefix])
        except FileNotFoundError as file_ex:
            raise DestinationUndefinedEntity(file_ex)

    def list_files_with_prefixes(self, table_dir: str, prefixes: List[str]) -> List[str]:
        """returns all files in a directory that match given prefixes"""
        result = []
        # we fallback to our own glob implementation that is tested to return consistent results for
        # filesystems we support. we were not able to use `find` or `walk` because they were selecting
        # files wrongly (on azure walk on path1/path2/ would also select files from path1/path2_v2/ but returning wrong dirs)
        for details in glob_files(self.fs_client, self.make_remote_url(table_dir), "**"):
            file = details["file_name"]
            filepath = self.pathlib.join(table_dir, details["relative_path"])
            # skip INIT files, do not remove: backward compat
            if file == INIT_FILE_NAME:
                continue
            for p in prefixes:
                if filepath.startswith(p):
                    result.append(filepath)
                    break
        return result

    def is_storage_initialized(self) -> bool:
        return self.fs_client.exists(self.init_file_path)  # type: ignore[no-any-return]

    @staticmethod
    def get_reference_followup_job_class(
        table: PreparedTableSchema,
    ) -> Optional[Type[ReferenceFollowupJob]]:
        # NOTE: we import deltalake/pyiceberg libs here to raise early if deps are not present
        if table.get("table_format") == "delta":
            import dlt.common.libs.deltalake

            return DeltaLoadFilesystemJob
        elif table.get("table_format") == "iceberg":
            import dlt.common.libs.pyiceberg

            return IcebergLoadFilesystemJob
        return None

    def get_load_job_class(self, table: PreparedTableSchema, file_path: str) -> Type[LoadJob]:
        if table["name"] == self.schema.state_table_name and not self.config.as_staging_destination:
            # skip the state table, we create a jsonl file in the complete_load step
            # this does not apply to scenarios where we are using filesystem as staging
            # where we want to load the state the regular way
            return FinalizedLoadJob
        elif reference_followup_job_class := self.get_reference_followup_job_class(table):
            if ReferenceFollowupJobRequest.is_reference_job(file_path):
                return reference_followup_job_class
            else:
                return FinalizedLoadJobWithFollowupJobs
        elif self.config.as_staging_destination:
            return FilesystemLoadJobWithFollowup
        else:
            return FilesystemLoadJob

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        return self.get_load_job_class(table, file_path)(file_path)

    def make_remote_url(self, remote_path: str) -> str:
        """Returns uri to the remote filesystem to which copy the file"""
        if self.is_local_filesystem:
            return self.config.make_file_url(remote_path)
        else:
            return self.config.make_url(remote_path)

    def __enter__(self) -> "FilesystemClient":
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        pass

    def should_load_data_to_staging_dataset(self, table_name: str) -> bool:
        return False

    def should_truncate_table_before_load(self, table_name: str) -> bool:
        table = self.prepare_load_table(table_name)
        return table["write_disposition"] == "replace" and not table.get("table_format") in (
            "delta",
            "iceberg",
        )  # Delta/Iceberg can do a logical replace

    #
    # state stuff
    #

    def _write_to_json_file(self, filepath: str, data: DictStrAny) -> None:
        self.fs_client.write_text(filepath, json.dumps(data), encoding="utf-8")

    def _to_path_safe_string(self, s: str) -> str:
        """for base64 strings"""
        return base64.b64decode(s).hex() if s else None

    def _list_dlt_table_files(
        self, table_name: str, pipeline_name: str = None
    ) -> Iterator[Tuple[str, List[str]]]:
        all_files = self.list_table_files(table_name)
        if len(all_files) == 0:
            if not self.is_storage_initialized():
                raise DestinationUndefinedEntity(table_name)
        for filepath in all_files:
            filename = os.path.splitext(os.path.basename(filepath))[0]
            fileparts = filename.rsplit(
                FILENAME_SEPARATOR, maxsplit=2
            )  # name, load_id, version_hash
            if len(fileparts) != 3:
                continue
            # Filters only if pipeline_name provided
            if pipeline_name is None or fileparts[0] == pipeline_name:
                yield filepath, fileparts

    def _store_load(self, load_id: str) -> None:
        # write entry to load "table"
        # TODO: this is also duplicate across all destinations. DRY this.
        load_data = {
            C_DLT_LOADS_TABLE_LOAD_ID: load_id,
            "schema_name": self.schema.name,
            "status": 0,
            "inserted_at": pendulum.now().isoformat(),
            "schema_version_hash": self.schema.version_hash,
        }
        filepath = self.pathlib.join(
            self.dataset_path,
            self.schema.loads_table_name,
            f"{self.schema.name}{FILENAME_SEPARATOR}{load_id}.jsonl",
        )
        self._write_to_json_file(filepath, load_data)

    def complete_load(self, load_id: str) -> None:
        # store current state
        self._store_current_state(load_id)
        self._store_load(load_id)

    #
    # state read/write
    #

    def _get_state_file_name(self, pipeline_name: str, version_hash: str, load_id: str) -> str:
        """gets full path for schema file for a given hash"""
        return self.pathlib.join(  # type: ignore[no-any-return]
            self.get_table_dir(self.schema.state_table_name),
            f"{pipeline_name}{FILENAME_SEPARATOR}{load_id}{FILENAME_SEPARATOR}{self._to_path_safe_string(version_hash)}.jsonl",
        )

    def _cleanup_pipeline_states(self, pipeline_name: str) -> None:
        state_table_files = list(
            self._list_dlt_table_files(self.schema.state_table_name, pipeline_name)
        )

        if len(state_table_files) > self.config.max_state_files:
            # filter and collect a list of state files
            state_file_info: List[Dict[str, Any]] = [
                {
                    "load_id": float(fileparts[1]),  # convert load_id to float for comparison
                    "filepath": filepath,
                }
                for filepath, fileparts in state_table_files
            ]

            # sort state file info by load_id in descending order
            state_file_info.sort(key=lambda x: x["load_id"], reverse=True)

            # keeping only the most recent MAX_STATE_HISTORY files
            files_to_delete = state_file_info[self.config.max_state_files :]

            # delete the old files
            for file_info in files_to_delete:
                self._delete_file(file_info["filepath"])

    def _store_current_state(self, load_id: str) -> None:
        # don't save the state this way when used as staging
        if self.config.as_staging_destination:
            return

        # get state doc from current pipeline
        try:
            pipeline_state_doc = current_load_package()["state"].get("pipeline_state")
        except CurrentLoadPackageStateNotAvailable:
            pipeline_state_doc = None

        if not pipeline_state_doc:
            return

        # get paths
        pipeline_name = pipeline_state_doc["pipeline_name"]
        hash_path = self._get_state_file_name(
            pipeline_name, self.schema.stored_version_hash, load_id
        )

        # write
        self._write_to_json_file(hash_path, cast(DictStrAny, pipeline_state_doc))

        # perform state cleanup only if max_state_files is set to a positive value
        if self.config.max_state_files >= 1:
            self._cleanup_pipeline_states(pipeline_name)

    def get_stored_state(self, pipeline_name: str) -> Optional[StateInfo]:
        # search newest state
        selected_path = None
        newest_load_id = "0"
        for filepath, fileparts in self._list_dlt_table_files(self.schema.state_table_name):
            if fileparts[0] == pipeline_name and fileparts[1] > newest_load_id:
                newest_load_id = fileparts[1]
                selected_path = filepath

        # Load compressed state from destination
        if selected_path:
            state_json: TPipelineStateDoc = json.loads(
                self.fs_client.read_text(selected_path, encoding="utf-8")
            )
            # we had dlt_load_id stored until version 0.5 and since we do not have any version control
            # we always migrate
            if load_id := state_json.pop("dlt_load_id", None):  # type: ignore[typeddict-item]
                state_json[C_DLT_LOAD_ID] = load_id  # type: ignore[literal-required]
            return StateInfo(**state_json)

        return None

    #
    # Schema read/write
    #

    def _get_schema_file_name(self, version_hash: str, load_id: str) -> str:
        """gets full path for schema file for a given hash"""

        return self.pathlib.join(  # type: ignore[no-any-return]
            self.get_table_dir(self.schema.version_table_name),
            f"{self.schema.name}{FILENAME_SEPARATOR}{load_id}{FILENAME_SEPARATOR}{self._to_path_safe_string(version_hash)}.jsonl",
        )

    def _iter_stored_schema_files(self) -> Iterator[Tuple[str, List[str]]]:
        """Iterator over all stored schema files"""
        for filepath, fileparts in self._list_dlt_table_files(self.schema.version_table_name):
            yield filepath, fileparts

    def _get_stored_schema_by_hash_or_newest(
        self, version_hash: str = None, schema_name: str = None
    ) -> Optional[StorageSchemaInfo]:
        """Get the schema by supplied hash, falls back to getting the newest version matching the existing schema name"""
        version_hash = self._to_path_safe_string(version_hash)
        # find newest schema for pipeline or by version hash
        try:
            selected_path = None
            newest_load_id = "0"
            for filepath, fileparts in self._iter_stored_schema_files():
                if (
                    not version_hash
                    and (fileparts[0] == schema_name or (not schema_name))
                    and fileparts[1] > newest_load_id
                ):
                    newest_load_id = fileparts[1]
                    selected_path = filepath
                elif fileparts[2] == version_hash:
                    selected_path = filepath
                    break

            if selected_path:
                info = json.loads(self.fs_client.read_text(selected_path, encoding="utf-8"))
                info["inserted_at"] = ensure_pendulum_datetime_utc(info["inserted_at"])
                return StorageSchemaInfo(**info)
        except DestinationUndefinedEntity:
            # ignore missing table
            pass

        return None

    def _update_schema_in_storage(self, schema: Schema) -> None:
        # get paths
        schema_id = str(time.precise_time())
        filepath = self._get_schema_file_name(schema.stored_version_hash, schema_id)

        # TODO: duplicate of weaviate implementation, should be abstracted out
        version_info = {
            "version_hash": schema.stored_version_hash,
            "schema_name": schema.name,
            "version": schema.version,
            "engine_version": schema.ENGINE_VERSION,
            "inserted_at": pendulum.now(),
            "schema": json.dumps(schema.to_dict()),
        }

        # we always keep tabs on what the current schema is
        self._write_to_json_file(filepath, version_info)

    def get_stored_schema(self, schema_name: str = None) -> Optional[StorageSchemaInfo]:
        """Retrieves newest schema from destination storage"""
        return self._get_stored_schema_by_hash_or_newest(schema_name=schema_name)

    def get_stored_schema_by_hash(self, version_hash: str) -> Optional[StorageSchemaInfo]:
        return self._get_stored_schema_by_hash_or_newest(version_hash)

    def create_table_chain_completed_followup_jobs(
        self,
        table_chain: Sequence[PreparedTableSchema],
        completed_table_chain_jobs: Optional[Sequence[LoadJobInfo]] = None,
    ) -> List[FollowupJobRequest]:
        assert completed_table_chain_jobs is not None
        jobs = super().create_table_chain_completed_followup_jobs(
            table_chain, completed_table_chain_jobs
        )
        if self.get_reference_followup_job_class(table_chain[0]):
            jobs.extend(
                self.get_reference_followup_job_requests(
                    table_chain,
                    completed_table_chain_jobs,
                )
            )
        return jobs

    def get_reference_followup_job_requests(
        self,
        table_chain: Sequence[PreparedTableSchema],
        completed_table_chain_jobs: Optional[Sequence[LoadJobInfo]],
    ) -> List["ReferenceFollowupJobRequest"]:
        def get_job_file_paths(*table_names: str) -> List[str]:
            return [
                job.file_path
                for job in completed_table_chain_jobs
                if job.job_file_info.table_name in table_names
            ]

        if self.batch_table_chain:
            job_file_paths = get_job_file_paths(*[t["name"] for t in table_chain])
            file_name = FileStorage.get_file_name_from_file_path(job_file_paths[0])
            return [ReferenceFollowupJobRequest(file_name, job_file_paths)]
        else:
            jobs = []
            for table in table_chain:
                job_file_paths = get_job_file_paths(table["name"])
                if job_file_paths:
                    file_name = FileStorage.get_file_name_from_file_path(job_file_paths[0])
                    jobs.append(ReferenceFollowupJobRequest(file_name, job_file_paths))
                else:
                    # file_name = ParsedLoadJobFileName(table["name"], "empty", 0, "reference").file_name()
                    # TODO: if we implement removal od orphaned rows, we may need to propagate such job without files
                    # to the delta load job
                    pass
            return jobs

    # SupportsOpenTables implementation

    def load_open_table(self, table_format: TTableFormat, table_name: str, **kwargs: Any) -> Any:
        """Locates, loads and returns native table client for table `table_name` in delta or iceberg formats"""

        try:
            prepared_table = self.prepare_load_table(table_name)
        except TableNotFound:
            raise DestinationUndefinedEntity(table_name)
        detected_format = prepared_table.get("table_format")
        if detected_format != table_format:
            raise OpenTableFormatNotSupported(table_format, table_name, detected_format)
        table_location = self.get_open_table_location(table_format, table_name)

        if table_format == "iceberg":
            catalog = self.get_open_table_catalog("iceberg")
            from dlt.common.libs.pyiceberg import evolve_table, NoSuchTableError

            table_id = f"{self.dataset_name}.{table_name}"
            try:
                return evolve_table(
                    catalog=catalog,
                    fs_client=self.fs_client,
                    config=self.config,
                    table_id=table_id,
                    table_location=table_location,
                    **kwargs,
                )
            except (NoSuchTableError, FileNotFoundError):
                raise DestinationUndefinedEntity(table_name)
        elif table_format == "delta":
            from dlt.common.libs.deltalake import deltalake_storage_options, DeltaTable

            storage_options = deltalake_storage_options(
                self.config.credentials, self.config.deltalake_storage_options
            )

            if not DeltaTable.is_deltatable(table_location, storage_options):
                raise DestinationUndefinedEntity(table_name)
            return DeltaTable(table_location, storage_options=storage_options)
        else:
            raise NotImplementedError(
                f"Can't load tables using `{table_format=:}` with `filesystem` destination."
            )

    def get_open_table_catalog(
        self, table_format: TTableFormat, catalog_name: Optional[str] = None
    ) -> Any:
        """Gets a native catalog for a table with format `table_format`

        Returns: currently pyiceberg Catalog is supported
        """
        if table_format != "iceberg":
            raise OpenTableCatalogNotSupported(table_format, "filesystem")

        if self._catalog:
            return self._catalog

        from dlt.common.libs.pyiceberg import get_catalog, IcebergCatalog
        from pyiceberg.exceptions import NamespaceAlreadyExistsError

        catalog: IcebergCatalog

        # Try to load catalog using new function
        catalog = self._catalog = get_catalog(
            iceberg_catalog_name=catalog_name or ConfigValue,
            credentials=self.config.credentials,
        )

        logger.info(f"Successfully loaded catalog '{catalog.name}' ")

        # Create namespace
        try:
            catalog.create_namespace(
                self.dataset_name,
                properties=self.config.iceberg_namespace_properties or {},
            )
            logger.info(f"Created Iceberg namespace: {self.dataset_name}")
        except NamespaceAlreadyExistsError as e:
            logger.debug(f"Namespace {self.dataset_name} already exists or error: {e}")

        return catalog

    def get_open_table_location(self, table_format: TTableFormat, table_name: str) -> str:
        """All tables have location, also those in "native" table format. Native format
        in case of filesystem is a set of parquet/csv/jsonl files where a table may
        be placed in a separate folder or share common prefix define in the layout.
        Locations of native tables will are normalized to include trailing separator
        if path is a "folder" (includes buckets)
        Note: location is fully formed url
        """
        prefix = self.get_table_prefix(table_name)
        location = self.make_remote_url(prefix)
        if self.config.is_local_filesystem and os.name == "nt":
            # pyiceberg cannot deal with windows absolute urls
            location = location.replace("file:///", "file://")
        return location

    def is_open_table(self, table_format: TTableFormat, table_name: str) -> bool:
        if table_name in self.schema.dlt_table_names():
            return False
        try:
            prepared_table = self.prepare_load_table(table_name)
        except TableNotFound:
            return False
        detected_format = prepared_table.get("table_format")
        return table_format == detected_format


# NOTE: HfFilesystemClient uses both fsspec and HfApi clients. We manually invalidate fsspec client
# cache after each filesystem mutation executed through HfApi to ensure consistency.
class HfFilesystemClient(FilesystemClient):
    batch_table_chain: ClassVar[bool] = True

    def __init__(
        self,
        schema: Schema,
        config: HfFilesystemDestinationClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        from huggingface_hub import HfApi

        super().__init__(schema, config, capabilities)
        self.config: HfFilesystemDestinationClientConfiguration = config
        self.hf_api = HfApi(**config.credentials.to_hf_api_credentials())

    @property
    def repo_id(self) -> str:
        return self.config.hf_namespace + "/" + self.dataset_name

    def create_dataset(self) -> None:
        self.hf_api.create_repo(repo_id=self.repo_id, repo_type="dataset")
        if self.config.hf_dataset_card:
            self._safe_card_operation("initialize dataset card", self.init_dataset_card)
        self.fs_client.invalidate_cache()

    def drop_dataset(self) -> None:
        self.hf_api.delete_repo(repo_id=self.repo_id, repo_type="dataset")
        self.fs_client.invalidate_cache()

    def complete_load(self, load_id: str) -> None:
        if self.config.hf_dataset_card:
            self._safe_card_operation(
                "update dataset card metadata", self.update_dataset_card_metadata, load_id
            )
        super().complete_load(load_id)

    def _collect_delete_operations(self, table_names: List[str]) -> List[Any]:
        """Collect CommitOperationDelete ops for files belonging to given tables."""
        from huggingface_hub import CommitOperationDelete

        ops: List[Any] = []
        table_dirs = set(self.get_table_dirs(table_names))
        table_prefixes = [self.get_table_prefix(t) for t in table_names]
        for table_dir in table_dirs:
            if self.fs_client.exists(table_dir):
                for file_path in self.list_files_with_prefixes(table_dir, table_prefixes):
                    path_in_repo = self.pathlib.relpath(file_path, self.dataset_path)
                    ops.append(CommitOperationDelete(path_in_repo=path_in_repo))
        return ops

    def _hf_commit_operations(self, operations: List[Any], commit_message: str) -> None:
        """Commit operations to HF repo and invalidate fsspec cache."""
        if not operations:
            return
        self.hf_api.create_commit(
            repo_id=self.repo_id,
            operations=operations,
            commit_message=commit_message,
            repo_type="dataset",
        )
        self.fs_client.invalidate_cache()

    def truncate_tables(self, table_names: List[str]) -> None:
        """Truncates tables in a single HF commit.
        Note: tables won't be present in storage but dlt schemas are not reset
        """
        ops = self._collect_delete_operations(table_names)
        self._hf_commit_operations(ops, f"Truncate tables: {', '.join(table_names)}")

    def drop_tables(self, *tables: str, delete_schema: bool = True) -> None:
        """Drop tables and optionally schema files in a single HF commit."""
        from huggingface_hub import CommitOperationDelete

        ops = self._collect_delete_operations(list(tables))
        if delete_schema:
            for filepath, fileparts in self._iter_stored_schema_files():
                if fileparts[0] == self.schema.name:
                    path_in_repo = self.pathlib.relpath(filepath, self.dataset_path)
                    ops.append(CommitOperationDelete(path_in_repo=path_in_repo))
        self._hf_commit_operations(ops, f"Drop tables: {', '.join(tables)}")

    def should_truncate_table_before_load(self, table_name: str) -> bool:
        # HF handles replace atomically in the commit job (delete + add in one commit)
        return False

    def get_load_job_class(self, table: PreparedTableSchema, file_path: str) -> Type[LoadJob]:
        if table["name"] == self.schema.state_table_name and not self.config.as_staging_destination:
            return FinalizedLoadJob
        elif ReferenceFollowupJobRequest.is_reference_job(file_path):
            return HfFilesystemCommitJob
        else:
            return HfFilesystemUploadJob

    def to_path_in_repo(self, path: str) -> str:
        """Strips dataset path prefix to return path relative to repo root."""
        return self.pathlib.relpath(path, self.dataset_path)  # type: ignore[no-any-return]

    def list_table_files_in_repo(self, table_name: str) -> List[str]:
        """Lists files for `table_name` with paths relative to repo root."""
        return [self.to_path_in_repo(path) for path in self.list_table_files(table_name)]

    @staticmethod
    def get_reference_followup_job_class(
        table: PreparedTableSchema,
    ) -> Type[HfFilesystemCommitJob]:
        return HfFilesystemCommitJob

    def _safe_card_operation(self, operation: str, func: Any, *args: Any) -> None:
        """Run a dataset card operation with retry and exception swallowing.

        Card operations are non-critical — a failure should not break data loads.
        """
        try:
            retry(
                wait=wait_exponential(multiplier=1, max=10),
                stop=stop_after_attempt(3),
                reraise=True,
            )(func)(*args)
        except Exception as ex:
            logger.warning(
                f"Failed to {operation} for {self.repo_id} after retries,"
                f" skipping: {type(ex).__name__}: {ex}"
            )

    @contextmanager
    def _hf_endpoint_env(self) -> Iterator[None]:
        """Temporarily sets HF_ENDPOINT env var as specified in credentials.

        Useful for methods that do not allow passing endpoint as argument.

        Behavior in all four cases:
        1. no endpoint in credentials, HF_ENDPOINT not set: no-op
        2. no endpoint in credentials, HF_ENDPOINT set: no-op
        3. endpoint in credentials, HF_ENDPOINT not set: sets HF_ENDPOINT to credentials endpoint temporarily
        4. endpoint in credentials, HF_ENDPOINT set: overrides HF_ENDPOINT with credentials endpoint temporarily
        """
        if endpoint := self.config.credentials.hf_endpoint:
            old = os.environ.get("HF_ENDPOINT")
            os.environ["HF_ENDPOINT"] = endpoint
            try:
                yield
            finally:
                if old is None:
                    del os.environ["HF_ENDPOINT"]
                else:
                    os.environ["HF_ENDPOINT"] = old
        else:
            yield

    def init_dataset_card(self) -> None:
        from huggingface_hub import DatasetCard

        with self._hf_endpoint_env():
            # TODO: make dataset card content configurable
            DatasetCard(content="").push_to_hub(
                repo_id=self.repo_id,
                token=self.config.credentials.hf_token,
                repo_type="dataset",
                commit_message="Initialize dataset card",
            )

    @staticmethod
    def create_dataset_card_metadata_config(
        config_name: str, data_file_paths: List[str]
    ) -> Dict[str, Any]:
        return {
            "config_name": config_name,
            "data_files": [{"split": "train", "path": data_file_paths}],
        }

    def create_dataset_card_metadata_configs(self) -> List[Dict[str, Any]]:
        return [
            self.create_dataset_card_metadata_config(
                config_name=table_name, data_file_paths=self.list_table_files_in_repo(table_name)
            )
            for table_name in self.schema.data_table_names(seen_data_only=True)
        ]

    def create_dataset_card_metadata(self) -> Dict[str, Any]:
        return {"configs": self.create_dataset_card_metadata_configs()}

    def update_dataset_card_metadata(self, load_id: str) -> None:
        from huggingface_hub import metadata_update

        table_names = list(self.schema.data_table_names(seen_data_only=True))
        with self._hf_endpoint_env():
            metadata_update(
                repo_id=self.repo_id,
                metadata=self.create_dataset_card_metadata(),
                repo_type="dataset",
                overwrite=True,
                token=self.config.credentials.hf_token,
                commit_message=(
                    f"Load {load_id}: update dataset card metadata for {', '.join(table_names)}"
                ),
            )
